import asyncio
import logging
import random
import time
from typing import AsyncGenerator

from apipeline.frames import (
    Frame,
    ErrorFrame,
    StartFrame,
    EndFrame,
    CancelFrame,
    AudioRawFrame,
    TextFrame,
)
from moshi.models.loaders import DEFAULT_REPO
import numpy as np
import torch

from achatbot.common.utils.audio_utils import (
    bytes2NpArrayWith16,
    postprocess_tts_wave_int16,
    resample_audio,
)
from achatbot.processors.voice.base import VoiceProcessorBase
from achatbot.types.llm.lmgen import LMGenArgs

try:
    from sphn import OpusStreamReader, OpusStreamWriter, resample
    from sentencepiece import SentencePieceProcessor
    from moshi.models import loaders, MimiModel, LMModel, LMGen
    from huggingface_hub import hf_hub_download
except ModuleNotFoundError as e:
    logging.error(f"Exception: {e}")
    logging.error(
        "In order to use moshi, you need to `pip install achatbot[moshi_voice_processor]`"
    )
    raise Exception(f"Missing module: {e}")


def seed_all(seed):
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)  # for multi-GPU setups
    random.seed(seed)
    np.random.seed(seed)
    torch.backends.cudnn.deterministic = False
    torch.backends.cudnn.benchmark = False


class MoshiVoiceBaseProcessor(VoiceProcessorBase):
    """
    use mimi speech codec + moshi lm(moshiko/moshika)
    - A1-T2A2: (speech)-to-(speech and text) (mimi(speech encoder)) -> (llm) -- text|speech tokens --> bpe(text decoder)|mimi(speech decoder))

    NOTE:
    - moshi genative lm no system prompt and funciton call.
    """

    def __init__(
        self,
        *,
        lm_gen_args: LMGenArgs | dict = LMGenArgs(),
        model_name: str = loaders.DEFAULT_REPO,
        mimi_weight_file: str | None = None,
        text_tokenizer_file: str | None = None,
        moshi_weight_file: str | None = None,
        device: str = "cuda",
        **kwargs,
    ):
        super().__init__(**kwargs)
        seed_all(12345678)
        self._lm_gen_args = lm_gen_args
        if isinstance(lm_gen_args, dict):
            self._lm_gen_args = LMGenArgs(**lm_gen_args)
        self._model_name = model_name
        self._mimi_weight_file = mimi_weight_file
        self._text_tokenizer_file = text_tokenizer_file
        self._moshi_weight_file = moshi_weight_file
        self._device = device

        self._mimi: MimiModel = None
        self._text_tokenizer: SentencePieceProcessor = None
        self._moshi_lm: LMModel = None
        self._lm_gen: LMGen = None

        self.load_models()

        # open streaming with batch_size:1,
        # if not, need use with lm_gen.streaming(1):
        self._mimi.streaming_forever(1)
        self._lm_gen.streaming_forever(1)
        self.warmup()

        self._cur_in_sample_rate = 0
        self._all_pcm_data = None

    @property
    def stream_info(self) -> dict:
        """Return dict out stream info"""
        return {
            "sample_rate": self._mimi.sample_rate,
            "channels": self._mimi.channels,
        }

    def load_models(self):
        logging.info("loading mimi")
        if self._mimi_weight_file is None:
            self._mimi_weight_file = hf_hub_download(self._model_name, loaders.MIMI_NAME)
        self._mimi = loaders.get_mimi(self._mimi_weight_file, self._device)
        self._mimi.set_num_codebooks(8)
        self._frame_size = int(self._mimi.sample_rate / self._mimi.frame_rate)
        logging.info("mimi loaded")

        logging.info("loading text tokenizer")
        if self._text_tokenizer_file is None:
            self._text_tokenizer_file = hf_hub_download(
                self._model_name, loaders.TEXT_TOKENIZER_NAME
            )
        self._text_tokenizer = SentencePieceProcessor(self._text_tokenizer_file)
        logging.info("loaded text tokenizer")

        logging.info("loading moshi")
        if self._moshi_weight_file is None:
            self._moshi_weight_file = hf_hub_download(self._model_name, loaders.MOSHI_NAME)
        self._moshi_lm = loaders.get_moshi_lm(self._moshi_weight_file, self._device)
        self._lm_gen = LMGen(
            self._moshi_lm,
            **self._lm_gen_args.__dict__,
        )
        logging.info("moshi loaded")

    def warmup(self):
        logging.info("start warmup")
        for i in range(4):
            be = time.time()
            chunk = torch.zeros(1, 1, self._frame_size, dtype=torch.float32, device=self._device)
            codes = self._mimi.encode(chunk)
            for c in range(codes.shape[-1]):
                tokens = self._lm_gen.step(codes[:, :, c : c + 1])
                if tokens is None:
                    continue
                _ = self._mimi.decode(tokens[:, 1:])
            logging.info(f"chunk warmup {i} in {1000 * (time.time() - be):.1f}ms")
        torch.cuda.synchronize()
        logging.info("end warmup")

    def resample(self, pcm):
        if self._cur_in_sample_rate != self._mimi.sample_rate:
            return resample(pcm, self._cur_in_sample_rate, self._mimi.sample_rate)
        return pcm


class MoshiVoiceOpusStreamProcessor(MoshiVoiceBaseProcessor):
    """
    use mimi speech codec + moshi lm(moshiko/moshika) with opus w/r streamer
    """

    def __init__(
        self,
        *,
        lm_gen_args: LMGenArgs | dict = LMGenArgs(),
        model_name: str = loaders.DEFAULT_REPO,
        mimi_weight_file: str | None = None,
        text_tokenizer_file: str | None = None,
        moshi_weight_file: str | None = None,
        device: str = "cuda",
        **kwargs,
    ):
        super().__init__(
            lm_gen_args=lm_gen_args,
            model_name=model_name,
            mimi_weight_file=mimi_weight_file,
            text_tokenizer_file=text_tokenizer_file,
            moshi_weight_file=moshi_weight_file,
            device=device,
            **kwargs,
        )

        self._opus_writer: OpusStreamWriter = None
        self._opus_reader: OpusStreamReader = None

        self._audio_in_task = None
        self._audio_out_task = None

    def reset_state(self):
        # https://opus-codec.org/
        # https://github.com/kyutai-labs/sphn (python binds rust lib so)
        # Easily load various audio file formats (bytes) into numpy arrays.
        # Read/write ogg/opus audio files with streaming support.
        # use Opus format for audio across the websocket,
        # as it can be safely streamed and decoded in real-time

        # opus writer encode pcm(float32) to apus format
        self._opus_writer = OpusStreamWriter(self._mimi.sample_rate)
        # NOTE:
        # new stream reader thread to decode apus format to pcm(float32),
        # https://github.com/kyutai-labs/sphn/blob/main/src/opus.rs#L345
        self._opus_reader = OpusStreamReader(self._mimi.sample_rate)

        # LLM is stateful, maintaining chat history,
        # so reset it on each connection
        self._mimi.reset_streaming()
        self._lm_gen.reset_streaming()

    async def start(self, frame: StartFrame):
        await super().start(frame)

        self.reset_state()

        self._audio_in_task = self.get_event_loop().create_task(self._audio_in_task_handler())
        self._audio_out_task = self.get_event_loop().create_task(self._audio_out_task_handler())

        self._create_push_task()

        logging.info("start done")

    async def stop(self, frame: EndFrame):
        if self._audio_in_task:
            self._audio_in_task.cancel()
            await self._audio_in_task
        if self._audio_out_task:
            self._audio_out_task.cancel()
            await self._audio_out_task

        self._opus_reader.close()

        await super().stop(frame)
        logging.info("stop done")

    async def cancel(self, frame: CancelFrame):
        await super().cancel(frame)
        logging.info("cancel done")

    async def _audio_in_task_handler(self):
        self._all_pcm_data = None
        while True:
            try:
                pcm = self._opus_reader.read_pcm()
                if pcm is None or pcm.shape[-1] == 0:
                    await asyncio.sleep(0.001)
                    continue

                pcm = self.resample(pcm)
                if self._all_pcm_data is None:
                    self._all_pcm_data = pcm
                else:
                    self._all_pcm_data = np.concatenate((self._all_pcm_data, pcm))

                await self.start_processing_metrics()
                ttfb_metric = True
                while self._all_pcm_data.shape[-1] >= self._frame_size:
                    be = time.time()

                    chunk = self._all_pcm_data[: self._frame_size]
                    self._all_pcm_data = self._all_pcm_data[self._frame_size :]

                    with torch.no_grad():
                        # chunk numpy ndArray -> chunk torch tensor
                        chunk = torch.from_numpy(chunk)
                        chunk = chunk.to(device=self._device)[None, None]

                        # mimi encode speech tensor chunk
                        codes = self._mimi.encode(chunk)

                        for c in range(codes.shape[-1]):
                            # lm gen tokens
                            ttfb_metric and await self.start_ttfb_metrics()
                            tokens = self._lm_gen.step(codes[:, :, c : c + 1])
                            ttfb_metric and await self.stop_ttfb_metrics()
                            ttfb_metric = False
                            if tokens is None:
                                continue

                            # mimi decode speech tensor tokens,
                            # append numpy ndArray tokens to apus writer
                            assert tokens.shape[1] == self._lm_gen.lm_model.dep_q + 1
                            main_pcm = self._mimi.decode(tokens[:, 1:])
                            main_pcm = main_pcm.cpu()
                            self._opus_writer.append_pcm(main_pcm[0, 0].numpy())
                            # print("_1___opus_writer.append_pcm____", main_pcm, main_pcm.shape)

                            # bpe decode text tensor tokens, replace bpe _
                            # send text frame
                            text_token = tokens[0, 0, 0].item()
                            if text_token not in (0, 3):
                                _text = self._text_tokenizer.id_to_piece(text_token)
                                _text = _text.replace("▁", " ")
                                logging.info(f"text token '{_text}'")
                                await self.push_frame(TextFrame(text=_text))

                    logging.info(f"frame handled in {1000 * (time.time() - be):.1f}ms")

                await self.stop_processing_metrics()
            except asyncio.CancelledError:
                break

    async def _audio_out_task_handler(self):
        while True:
            try:
                # TODO: check read why slow
                audio_bytes = self._opus_writer.read_bytes()
                # print("_2__opus_writer.read_bytes____", audio_bytes)
                if not audio_bytes or len(audio_bytes) == 0:
                    await asyncio.sleep(0.001)
                    continue

                await self.queue_frame(
                    AudioRawFrame(
                        audio=audio_bytes,
                        sample_rate=self._mimi.sample_rate,
                        num_channels=self._mimi.channels,
                    )
                )
            except asyncio.CancelledError:
                break

    async def run_voice(self, frame: AudioRawFrame) -> AsyncGenerator[Frame, None]:
        """
        StreamReader sample_rate: 48000, 24000
        https://github.com/kyutai-labs/sphn/blob/main/src/opus.rs#L337
        """
        self._cur_in_sample_rate = frame.sample_rate
        self._opus_reader.append_bytes(frame.audio)
        yield None


class MoshiVoiceProcessor(MoshiVoiceBaseProcessor):
    async def run_voice(self, frame: AudioRawFrame) -> AsyncGenerator[Frame, None]:
        """
        StreamReader sample_rate: 48000, 24000
        https://github.com/kyutai-labs/sphn/blob/main/src/opus.rs#L337
        !TODO: need buffer
        """
        self._cur_in_sample_rate = frame.sample_rate
        pcm = bytes2NpArrayWith16(frame.audio)
        pcm = self.resample(pcm)
        if pcm.shape[-1] == 0:
            yield None
            return
        if self._all_pcm_data is None:
            self._all_pcm_data = pcm
        else:
            self._all_pcm_data = np.concatenate((self._all_pcm_data, pcm))
        await self.start_processing_metrics()
        ttfb_metric = True
        while self._all_pcm_data.shape[-1] >= self._frame_size:
            be = time.time()
            chunk = self._all_pcm_data[: self._frame_size]
            self._all_pcm_data = self._all_pcm_data[self._frame_size :]
            chunk = torch.from_numpy(chunk)
            chunk = chunk.to(device=self._device)[None, None]
            codes = self._mimi.encode(chunk)
            for c in range(codes.shape[-1]):
                ttfb_metric and await self.start_ttfb_metrics()
                tokens = self._lm_gen.step(codes[:, :, c : c + 1])
                ttfb_metric and await self.stop_ttfb_metrics()
                ttfb_metric = False
                if tokens is None:
                    continue
                assert tokens.shape[1] == self._lm_gen.lm_model.dep_q + 1
                main_pcm = self._mimi.decode(tokens[:, 1:])
                main_pcm = main_pcm.cpu()
                audio_bytes = postprocess_tts_wave_int16(main_pcm[0, 0])
                yield AudioRawFrame(
                    audio=audio_bytes,
                    sample_rate=self._mimi.sample_rate,
                    num_channels=self._mimi.channels,
                )
                text_token = tokens[0, 0, 0].item()
                if text_token not in (0, 3):
                    _text = self._text_tokenizer.id_to_piece(text_token)
                    _text = _text.replace("▁", " ")
                    logging.debug(f"text token '{_text}'")
                    yield TextFrame(text=_text)
            logging.info(f"frame handled in {1000 * (time.time() - be):.1f}ms")
        await self.stop_processing_metrics()


class VoiceOpusStreamEchoProcessor(VoiceProcessorBase):
    """
    use opuse stream to echo
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        seed_all(12345678)

        self.sample_rate = 24000
        self.channels = 1
        self._frame_size = 1920

        self._cur_in_sample_rate = 0
        self._all_pcm_data = None

        self._opus_writer: OpusStreamWriter = None
        self._opus_reader: OpusStreamReader = None

        self._audio_in_task = None
        self._audio_out_task = None

    @property
    def stream_info(self) -> dict:
        """Return dict out stream info"""
        return {
            "sample_rate": self.sample_rate,
            "channels": self.channels,
        }

    def resample(self, pcm):
        if self._cur_in_sample_rate != self.sample_rate:
            print(self._cur_in_sample_rate, "->", self.sample_rate)
            return resample(pcm, self._cur_in_sample_rate, self.sample_rate)
        return pcm

    def reset_state(self):
        # https://opus-codec.org/
        # https://github.com/kyutai-labs/sphn (python binds rust lib so)
        # Easily load various audio file formats (bytes) into numpy arrays.
        # Read/write ogg/opus audio files with streaming support.
        # use Opus format for audio across the websocket,
        # as it can be safely streamed and decoded in real-time

        # opus writer encode pcm(float32) to apus format
        self._opus_writer = OpusStreamWriter(self.sample_rate)
        # NOTE:
        # new stream reader thread to decode apus format to pcm(float32),
        # https://github.com/kyutai-labs/sphn/blob/main/src/opus.rs#L345
        self._opus_reader = OpusStreamReader(self.sample_rate)

    async def start(self, frame: StartFrame):
        await super().start(frame)

        self.reset_state()

        self._audio_in_task = self.get_event_loop().create_task(self._audio_in_task_handler())
        self._audio_out_task = self.get_event_loop().create_task(self._audio_out_task_handler())

        self._create_push_task()

        logging.info("start done")

    async def stop(self, frame: EndFrame):
        if self._audio_in_task:
            self._audio_in_task.cancel()
            await self._audio_in_task
        if self._audio_out_task:
            self._audio_out_task.cancel()
            await self._audio_out_task

        self._opus_reader.close()

        await super().stop(frame)
        logging.info("stop done")

    async def cancel(self, frame: CancelFrame):
        await super().cancel(frame)
        logging.info("cancel done")

    async def _audio_in_task_handler(self):
        self._all_pcm_data = None
        while True:
            try:
                pcm = self._opus_reader.read_pcm()
                # print(pcm, type(pcm), pcm.shape)
                if pcm.shape[-1] == 0:
                    # sleep to yield to avoid busy loop
                    await asyncio.sleep(0.001)
                    continue
                pcm = self.resample(pcm)
                if self._all_pcm_data is None:
                    self._all_pcm_data = pcm
                else:
                    self._all_pcm_data = np.concatenate((self._all_pcm_data, pcm))
                await self.start_processing_metrics()
                while self._all_pcm_data.shape[-1] >= self._frame_size:
                    be = time.time()
                    chunk = self._all_pcm_data[: self._frame_size]
                    self._all_pcm_data = self._all_pcm_data[self._frame_size :]
                    # print("1", chunk, type(chunk), chunk.shape)
                    self._opus_writer.append_pcm(chunk)
                    logging.debug(f"frame handled in {1000 * (time.time() - be):.1f}ms")
                await self.stop_processing_metrics()
            except asyncio.CancelledError:
                break

    async def _audio_out_task_handler(self):
        while True:
            try:
                await asyncio.sleep(0.001)
                audio_bytes = self._opus_writer.read_bytes()
                if len(audio_bytes) > 0:
                    # print("audio_bytes====>", len(audio_bytes))
                    await self.queue_frame(
                        AudioRawFrame(
                            audio=audio_bytes,
                            sample_rate=self.sample_rate,
                            num_channels=self.channels,
                        )
                    )
            except asyncio.CancelledError:
                break

    async def run_voice(self, frame: AudioRawFrame) -> AsyncGenerator[Frame, None]:
        """
        StreamReader sample_rate: 48000, 24000
        https://github.com/kyutai-labs/sphn/blob/main/src/opus.rs#L337
        """
        self._cur_in_sample_rate = frame.sample_rate
        # print("---frame---", frame)
        self._opus_reader.append_bytes(frame.audio)
        yield None



import os
import logging

import unittest
from dotenv import load_dotenv

from src.core.llm.personalai import PersonalAIProxy
from src.common.logger import Logger
from src.common.session import Session
from src.common.types import SessionCtx
from src.core.llm import LLMEnvInit

load_dotenv(override=True)

r"""
API_URL=http://localhost:8787/ \
    python -m unittest test.core.llm.test_personalai.TestPersonalAIProxy.test_chat_completion
API_URL=http://localhost:8787/ \
    CHAT_TYPE=chat_with_functions \
    LLM_PROMPT=今天北京天气怎么样 \
    python -m unittest test.core.llm.test_personalai.TestPersonalAIProxy.test_chat_completion

API_URL=http://localhost:8787/ \
    CHAT_BOT=qianfan LLM_MODEL_NAME=completions \
    python -m unittest test.core.llm.test_personalai.TestPersonalAIProxy.test_chat_completion
API_URL=http://localhost:8787/ \
    CHAT_BOT=qianfan LLM_MODEL_NAME=completions \
    CHAT_TYPE=chat_with_functions \
    LLM_PROMPT=今天北京天气怎么样 \
    python -m unittest test.core.llm.test_personalai.TestPersonalAIProxy.test_chat_completion

API_URL=https://personal-ai-ts.weedge.workers.dev/ \
    CHAT_BOT=openai \
    python -m unittest test.core.llm.test_personalai.TestPersonalAIProxy.test_chat_completion
API_URL=https://personal-ai-ts.weedge.workers.dev/ \
    CHAT_BOT=openai \
    CHAT_TYPE=chat_with_functions \
    LLM_PROMPT=今天北京天气怎么样 \
    python -m unittest test.core.llm.test_personalai.TestPersonalAIProxy.test_chat_completion

API_URL=https://personal-ai-ts.weedge.workers.dev/ \
    CHAT_BOT=qianfan LLM_MODEL_NAME=completions \
    python -m unittest test.core.llm.test_personalai.TestPersonalAIProxy.test_chat_completion
API_URL=https://personal-ai-ts.weedge.workers.dev/ \
    CHAT_BOT=qianfan LLM_MODEL_NAME=completions\
    CHAT_TYPE=chat_with_functions \
    LLM_PROMPT=今天北京天气怎么样 \
    python -m unittest test.core.llm.test_personalai.TestPersonalAIProxy.test_chat_completion
"""


class TestPersonalAIProxy(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.llm_tag = os.getenv("LLM_TAG", "llm_personalai_proxy")
        cls.prompt = os.getenv("LLM_PROMPT", "你好")
        Logger.init(os.getenv("LOG_LEVEL", "debug").upper(), is_file=False)

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        kwargs = LLMEnvInit.map_config_func[self.llm_tag]()
        self.engine: PersonalAIProxy = PersonalAIProxy(**kwargs)
        logging.info(f"initLLMEngine: {self.llm_tag}, {self.engine}")
        self.session = Session(**SessionCtx(f"test_{self.llm_tag}_client_id").__dict__)

    def tearDown(self):
        pass

    def test_chat_completion(self):
        self.engine.args.llm_chat_system = "你是一个中国人,一名中文助理，请用中文简短回答，回答限制在1-5句话内。要友好、乐于助人且简明扼要。保持对话简短而甜蜜。只用纯文本回答，不要包含链接或其他附加内容。不要回复计算机代码以及数学公式。"
        self.session.ctx.state["prompt"] = self.prompt
        logging.debug(self.session.ctx)
        iter = self.engine.chat_completion(self.session)
        for item in iter:
            print(item)
            self.assertGreater(len(item), 0)



from lightning.pytorch.utilities import rank_zero_only

from fish_speech.utils import logger as log


@rank_zero_only
def log_hyperparameters(object_dict: dict) -> None:
    """Controls which config parts are saved by lightning loggers.

    Additionally saves:
    - Number of model parameters
    """

    hparams = {}

    cfg = object_dict["cfg"]
    model = object_dict["model"]
    trainer = object_dict["trainer"]

    if not trainer.logger:
        log.warning("Logger not found! Skipping hyperparameter logging...")
        return

    hparams["model"] = cfg["model"]

    # save number of model parameters
    hparams["model/params/total"] = sum(p.numel() for p in model.parameters())
    hparams["model/params/trainable"] = sum(
        p.numel() for p in model.parameters() if p.requires_grad
    )
    hparams["model/params/non_trainable"] = sum(
        p.numel() for p in model.parameters() if not p.requires_grad
    )

    hparams["data"] = cfg["data"]
    hparams["trainer"] = cfg["trainer"]

    hparams["callbacks"] = cfg.get("callbacks")
    hparams["extras"] = cfg.get("extras")

    hparams["task_name"] = cfg.get("task_name")
    hparams["tags"] = cfg.get("tags")
    hparams["ckpt_path"] = cfg.get("ckpt_path")
    hparams["seed"] = cfg.get("seed")

    # send hparams to all loggers
    for logger in trainer.loggers:
        logger.log_hyperparams(hparams)



import os
import json
import logging
import http.client
from typing import List

import typer
from dotenv import load_dotenv

# Load environment variables from .env file
load_dotenv(override=True)

app = typer.Typer()


@app.command("d1_table_query")
def d1_table_query(db_id: str, sql: str, sql_params: List[str] = []) -> dict:
    """
    https://developers.cloudflare.com/api/operations/cloudflare-d1-query-database
    """
    account_id = os.getenv("CLOUDFLARE_ACCOUNT_ID")
    api_key = os.getenv("CLOUDFLARE_API_KEY")

    payload = {
        "params": sql_params,
        "sql": sql,
    }
    body = json.dumps(payload)
    headers = {
        "Authorization": f"Bearer {api_key}",
        "Content-Type": "application/json",
    }

    conn = http.client.HTTPSConnection("api.cloudflare.com")
    conn.request(
        "POST",
        f"/client/v4/accounts/{account_id}/d1/database/{db_id}/query",
        body,
        headers,
    )
    res = conn.getresponse()
    data = res.read().decode("utf-8")
    # print(data)
    json_data = json.loads(data)
    logging.info(f"body:{body}, db_id:{db_id}, query res:{json_data}")
    return json_data


@app.command("d1_db")
def d1_db(db_id: str) -> dict:
    """
    https://developers.cloudflare.com/api/operations/cloudflare-d1-get-database
    """
    account_id = os.getenv("CLOUDFLARE_ACCOUNT_ID")
    api_key = os.getenv("CLOUDFLARE_API_KEY")

    conn = http.client.HTTPSConnection("api.cloudflare.com")
    headers = {
        "Authorization": f"Bearer {api_key}",
        "Content-Type": "application/json",
    }

    conn.request(
        "GET",
        f"/client/v4/accounts/{account_id}/d1/database/{db_id}",
        headers=headers,
    )

    res = conn.getresponse()
    data = res.read()

    data = res.read().decode("utf-8")
    json_data = json.loads(data)
    logging.info(f"get db_id:{db_id}, query res:{json_data}")
    return json_data


r"""
python -m demo.cloudflare.rest_api d1_db \
    09f7a7c7-66ae-41ea-9dbe-b8b635b19758

python -m demo.cloudflare.rest_api d1_table_query \
    09f7a7c7-66ae-41ea-9dbe-b8b635b19758 \
    "select * from podcast limit 1"
"""
if __name__ == "__main__":
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s - %(name)s - %(levelname)s - %(pathname)s:%(lineno)d - %(funcName)s - %(message)s",
        handlers=[logging.StreamHandler()],
    )
    app()



r"""
The file creates a pickle file where the values needed for loading of dataset is stored and the model can load it
when needed.

Parameters from hparam.py will be used
"""
import argparse
import json
import os
import sys
from pathlib import Path

import rootutils
import torch
from hydra import compose, initialize
from omegaconf import open_dict
from tqdm.auto import tqdm

from matcha.data.text_mel_datamodule import TextMelDataModule
from matcha.utils.logging_utils import pylogger

log = pylogger.get_pylogger(__name__)


def compute_data_statistics(data_loader: torch.utils.data.DataLoader, out_channels: int):
    """Generate data mean and standard deviation helpful in data normalisation

    Args:
        data_loader (torch.utils.data.Dataloader): _description_
        out_channels (int): mel spectrogram channels
    """
    total_mel_sum = 0
    total_mel_sq_sum = 0
    total_mel_len = 0

    for batch in tqdm(data_loader, leave=False):
        mels = batch["y"]
        mel_lengths = batch["y_lengths"]

        total_mel_len += torch.sum(mel_lengths)
        total_mel_sum += torch.sum(mels)
        total_mel_sq_sum += torch.sum(torch.pow(mels, 2))

    data_mean = total_mel_sum / (total_mel_len * out_channels)
    data_std = torch.sqrt((total_mel_sq_sum / (total_mel_len * out_channels)) - torch.pow(data_mean, 2))

    return {"mel_mean": data_mean.item(), "mel_std": data_std.item()}


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-i",
        "--input-config",
        type=str,
        default="vctk.yaml",
        help="The name of the yaml config file under configs/data",
    )

    parser.add_argument(
        "-b",
        "--batch-size",
        type=int,
        default="256",
        help="Can have increased batch size for faster computation",
    )

    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        default=False,
        required=False,
        help="force overwrite the file",
    )
    args = parser.parse_args()
    output_file = Path(args.input_config).with_suffix(".json")

    if os.path.exists(output_file) and not args.force:
        print("File already exists. Use -f to force overwrite")
        sys.exit(1)

    with initialize(version_base="1.3", config_path="../../configs/data"):
        cfg = compose(config_name=args.input_config, return_hydra_config=True, overrides=[])

    root_path = rootutils.find_root(search_from=__file__, indicator=".project-root")

    with open_dict(cfg):
        del cfg["hydra"]
        del cfg["_target_"]
        cfg["data_statistics"] = None
        cfg["seed"] = 1234
        cfg["batch_size"] = args.batch_size
        cfg["train_filelist_path"] = str(os.path.join(root_path, cfg["train_filelist_path"]))
        cfg["valid_filelist_path"] = str(os.path.join(root_path, cfg["valid_filelist_path"]))

    text_mel_datamodule = TextMelDataModule(**cfg)
    text_mel_datamodule.setup()
    data_loader = text_mel_datamodule.train_dataloader()
    log.info("Dataloader loaded! Now computing stats...")
    params = compute_data_statistics(data_loader, cfg["n_feats"])
    print(params)
    json.dump(
        params,
        open(output_file, "w"),
    )


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""基本类
中文字符类
中文数字/数位类
中文数字类
中文数位类
中文数字系统类
中文数学符号类
*中文其他符号类
"""

__author__ = "Zhiyang Zhou <zyzhou@stu.xmu.edu.cn>"
__data__ = "2019-05-02"

from fish_speech.text.chn_text_norm.basic_constant import NUMBERING_TYPES


class ChineseChar(object):
    """
    中文字符
    每个字符对应简体和繁体,
    e.g. 简体 = '负', 繁体 = '負'
    转换时可转换为简体或繁体
    """

    def __init__(self, simplified, traditional):
        self.simplified = simplified
        self.traditional = traditional
        self.__repr__ = self.__str__

    def __str__(self):
        return self.simplified or self.traditional or None

    def __repr__(self):
        return self.__str__()


class ChineseNumberUnit(ChineseChar):
    """
    中文数字/数位字符
    每个字符除繁简体外还有一个额外的大写字符
    e.g. '陆' 和 '陸'
    """

    def __init__(self, power, simplified, traditional, big_s, big_t):
        super(ChineseNumberUnit, self).__init__(simplified, traditional)
        self.power = power
        self.big_s = big_s
        self.big_t = big_t

    def __str__(self):
        return "10^{}".format(self.power)

    @classmethod
    def create(cls, index, value, numbering_type=NUMBERING_TYPES[1], small_unit=False):

        if small_unit:
            return ChineseNumberUnit(
                power=index + 1,
                simplified=value[0],
                traditional=value[1],
                big_s=value[1],
                big_t=value[1],
            )
        elif numbering_type == NUMBERING_TYPES[0]:
            return ChineseNumberUnit(
                power=index + 8,
                simplified=value[0],
                traditional=value[1],
                big_s=value[0],
                big_t=value[1],
            )
        elif numbering_type == NUMBERING_TYPES[1]:
            return ChineseNumberUnit(
                power=(index + 2) * 4,
                simplified=value[0],
                traditional=value[1],
                big_s=value[0],
                big_t=value[1],
            )
        elif numbering_type == NUMBERING_TYPES[2]:
            return ChineseNumberUnit(
                power=pow(2, index + 3),
                simplified=value[0],
                traditional=value[1],
                big_s=value[0],
                big_t=value[1],
            )
        else:
            raise ValueError(
                "Counting type should be in {0} ({1} provided).".format(
                    NUMBERING_TYPES, numbering_type
                )
            )


class ChineseNumberDigit(ChineseChar):
    """
    中文数字字符
    """

    def __init__(
        self, value, simplified, traditional, big_s, big_t, alt_s=None, alt_t=None
    ):
        super(ChineseNumberDigit, self).__init__(simplified, traditional)
        self.value = value
        self.big_s = big_s
        self.big_t = big_t
        self.alt_s = alt_s
        self.alt_t = alt_t

    def __str__(self):
        return str(self.value)

    @classmethod
    def create(cls, i, v):
        return ChineseNumberDigit(i, v[0], v[1], v[2], v[3])


class ChineseMath(ChineseChar):
    """
    中文数位字符
    """

    def __init__(self, simplified, traditional, symbol, expression=None):
        super(ChineseMath, self).__init__(simplified, traditional)
        self.symbol = symbol
        self.expression = expression
        self.big_s = simplified
        self.big_t = traditional


CC, CNU, CND, CM = ChineseChar, ChineseNumberUnit, ChineseNumberDigit, ChineseMath


class NumberSystem(object):
    """
    中文数字系统
    """

    pass


class MathSymbol(object):
    """
    用于中文数字系统的数学符号 (繁/简体), e.g.
    positive = ['正', '正']
    negative = ['负', '負']
    point = ['点', '點']
    """

    def __init__(self, positive, negative, point):
        self.positive = positive
        self.negative = negative
        self.point = point

    def __iter__(self):
        for v in self.__dict__.values():
            yield v


# class OtherSymbol(object):
#     """
#     其他符号
#     """
#
#     def __init__(self, sil):
#         self.sil = sil
#
#     def __iter__(self):
#         for v in self.__dict__.values():
#             yield v



from dataclasses import dataclass, field
import os

from src.common.types import MODELS_DIR
from src.types.codec import CodecArgs
from src.types.llm.transformers import TransformersSpeechLMArgs


@dataclass
class LlasaTTSArgs:
    """
    TransformersManualSpeechLlasa LM + xcodec2 -> llasa TTS
    """

    lm_args: dict = field(default_factory=TransformersSpeechLMArgs().__dict__)
    xcode2_args: dict = field(default_factory=CodecArgs().__dict__)

    # ref audio file path and prompt text
    ref_audio_file_path: str = ""
    prompt_text: str = ""
    # is save ref audio encode vq code indices .npy
    is_save: bool = False
    output_codebook_indices_dir: str = os.path.join(MODELS_DIR, "llasa_codebook_indices")



from concurrent.futures import ThreadPoolExecutor
from queue import Queue
from typing import Generator

from src.common.session import Session
from src.common.factory import EngineClass
from src.types.speech.asr.base import ASRArgs
from src.common.utils import task
from src.common.interface import IAsr


class ASRBase(EngineClass, IAsr):
    @classmethod
    def get_args(cls, **kwargs) -> dict:
        return {**ASRArgs().__dict__, **kwargs}

    def __init__(self, **args) -> None:
        self.args = ASRArgs(**args)
        self.asr_audio = None

    def transcribe_stream_sync(self, session: Session) -> Generator[str, None, None]:
        queue: Queue = Queue()
        with ThreadPoolExecutor() as executor:
            executor.submit(task.fetch_async_items, queue, self.transcribe_stream, session)
            while True:
                item = queue.get()
                if item is None:
                    break
                yield item



# Copyright (c) 2024 Alibaba Inc (authors: Xiang Lyu, Kai Hu)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""HIFI-GAN"""

from typing import Dict, Optional, List
import numpy as np
from scipy.signal import get_window
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn import Conv1d
from torch.nn import ConvTranspose1d
from torch.nn.utils import remove_weight_norm
from torch.nn.utils.parametrizations import weight_norm
from torch.distributions.uniform import Uniform

from cosyvoice.transformer.activation import Snake
from cosyvoice.utils.common import get_padding
from cosyvoice.utils.common import init_weights


"""hifigan based generator implementation.

This code is modified from https://github.com/jik876/hifi-gan
 ,https://github.com/kan-bayashi/ParallelWaveGAN and
 https://github.com/NVIDIA/BigVGAN

"""


class ResBlock(torch.nn.Module):
    """Residual block module in HiFiGAN/BigVGAN."""
    def __init__(
        self,
        channels: int = 512,
        kernel_size: int = 3,
        dilations: List[int] = [1, 3, 5],
    ):
        super(ResBlock, self).__init__()
        self.convs1 = nn.ModuleList()
        self.convs2 = nn.ModuleList()

        for dilation in dilations:
            self.convs1.append(
                weight_norm(
                    Conv1d(
                        channels,
                        channels,
                        kernel_size,
                        1,
                        dilation=dilation,
                        padding=get_padding(kernel_size, dilation)
                    )
                )
            )
            self.convs2.append(
                weight_norm(
                    Conv1d(
                        channels,
                        channels,
                        kernel_size,
                        1,
                        dilation=1,
                        padding=get_padding(kernel_size, 1)
                    )
                )
            )
        self.convs1.apply(init_weights)
        self.convs2.apply(init_weights)
        self.activations1 = nn.ModuleList([
            Snake(channels, alpha_logscale=False)
            for _ in range(len(self.convs1))
        ])
        self.activations2 = nn.ModuleList([
            Snake(channels, alpha_logscale=False)
            for _ in range(len(self.convs2))
        ])

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        for idx in range(len(self.convs1)):
            xt = self.activations1[idx](x)
            xt = self.convs1[idx](xt)
            xt = self.activations2[idx](xt)
            xt = self.convs2[idx](xt)
            x = xt + x
        return x

    def remove_weight_norm(self):
        for idx in range(len(self.convs1)):
            remove_weight_norm(self.convs1[idx])
            remove_weight_norm(self.convs2[idx])


class SineGen(torch.nn.Module):
    """ Definition of sine generator
    SineGen(samp_rate, harmonic_num = 0,
            sine_amp = 0.1, noise_std = 0.003,
            voiced_threshold = 0,
            flag_for_pulse=False)
    samp_rate: sampling rate in Hz
    harmonic_num: number of harmonic overtones (default 0)
    sine_amp: amplitude of sine-wavefrom (default 0.1)
    noise_std: std of Gaussian noise (default 0.003)
    voiced_thoreshold: F0 threshold for U/V classification (default 0)
    flag_for_pulse: this SinGen is used inside PulseGen (default False)
    Note: when flag_for_pulse is True, the first time step of a voiced
        segment is always sin(np.pi) or cos(0)
    """

    def __init__(self, samp_rate, harmonic_num=0,
                 sine_amp=0.1, noise_std=0.003,
                 voiced_threshold=0):
        super(SineGen, self).__init__()
        self.sine_amp = sine_amp
        self.noise_std = noise_std
        self.harmonic_num = harmonic_num
        self.sampling_rate = samp_rate
        self.voiced_threshold = voiced_threshold

    def _f02uv(self, f0):
        # generate uv signal
        uv = (f0 > self.voiced_threshold).type(torch.float32)
        return uv

    @torch.no_grad()
    def forward(self, f0):
        """
        :param f0: [B, 1, sample_len], Hz
        :return: [B, 1, sample_len]
        """

        F_mat = torch.zeros((f0.size(0), self.harmonic_num + 1, f0.size(-1))).to(f0.device)
        for i in range(self.harmonic_num + 1):
            F_mat[:, i: i + 1, :] = f0 * (i + 1) / self.sampling_rate

        theta_mat = 2 * np.pi * (torch.cumsum(F_mat, dim=-1) % 1)
        u_dist = Uniform(low=-np.pi, high=np.pi)
        phase_vec = u_dist.sample(sample_shape=(f0.size(0), self.harmonic_num + 1, 1)).to(F_mat.device)
        phase_vec[:, 0, :] = 0

        # generate sine waveforms
        sine_waves = self.sine_amp * torch.sin(theta_mat + phase_vec)

        # generate uv signal
        uv = self._f02uv(f0)

        # noise: for unvoiced should be similar to sine_amp
        #        std = self.sine_amp/3 -> max value ~ self.sine_amp
        # .       for voiced regions is self.noise_std
        noise_amp = uv * self.noise_std + (1 - uv) * self.sine_amp / 3
        noise = noise_amp * torch.randn_like(sine_waves)

        # first: set the unvoiced part to 0 by uv
        # then: additive noise
        sine_waves = sine_waves * uv + noise
        return sine_waves, uv, noise


class SourceModuleHnNSF(torch.nn.Module):
    """ SourceModule for hn-nsf
    SourceModule(sampling_rate, harmonic_num=0, sine_amp=0.1,
                 add_noise_std=0.003, voiced_threshod=0)
    sampling_rate: sampling_rate in Hz
    harmonic_num: number of harmonic above F0 (default: 0)
    sine_amp: amplitude of sine source signal (default: 0.1)
    add_noise_std: std of additive Gaussian noise (default: 0.003)
        note that amplitude of noise in unvoiced is decided
        by sine_amp
    voiced_threshold: threhold to set U/V given F0 (default: 0)
    Sine_source, noise_source = SourceModuleHnNSF(F0_sampled)
    F0_sampled (batchsize, length, 1)
    Sine_source (batchsize, length, 1)
    noise_source (batchsize, length 1)
    uv (batchsize, length, 1)
    """

    def __init__(self, sampling_rate, upsample_scale, harmonic_num=0, sine_amp=0.1,
                 add_noise_std=0.003, voiced_threshod=0):
        super(SourceModuleHnNSF, self).__init__()

        self.sine_amp = sine_amp
        self.noise_std = add_noise_std

        # to produce sine waveforms
        self.l_sin_gen = SineGen(sampling_rate, harmonic_num,
                                 sine_amp, add_noise_std, voiced_threshod)

        # to merge source harmonics into a single excitation
        self.l_linear = torch.nn.Linear(harmonic_num + 1, 1)
        self.l_tanh = torch.nn.Tanh()

    def forward(self, x):
        """
        Sine_source, noise_source = SourceModuleHnNSF(F0_sampled)
        F0_sampled (batchsize, length, 1)
        Sine_source (batchsize, length, 1)
        noise_source (batchsize, length 1)
        """
        # source for harmonic branch
        with torch.no_grad():
            sine_wavs, uv, _ = self.l_sin_gen(x.transpose(1, 2))
            sine_wavs = sine_wavs.transpose(1, 2)
            uv = uv.transpose(1, 2)
        sine_merge = self.l_tanh(self.l_linear(sine_wavs))

        # source for noise branch, in the same shape as uv
        noise = torch.randn_like(uv) * self.sine_amp / 3
        return sine_merge, noise, uv


class HiFTGenerator(nn.Module):
    """
    HiFTNet Generator: Neural Source Filter + ISTFTNet
    https://arxiv.org/abs/2309.09493
    """
    def __init__(
            self,
            in_channels: int = 80,
            base_channels: int = 512,
            nb_harmonics: int = 8,
            sampling_rate: int = 22050,
            nsf_alpha: float = 0.1,
            nsf_sigma: float = 0.003,
            nsf_voiced_threshold: float = 10,
            upsample_rates: List[int] = [8, 8],
            upsample_kernel_sizes: List[int] = [16, 16],
            istft_params: Dict[str, int] = {"n_fft": 16, "hop_len": 4},
            resblock_kernel_sizes: List[int] = [3, 7, 11],
            resblock_dilation_sizes: List[List[int]] = [[1, 3, 5], [1, 3, 5], [1, 3, 5]],
            source_resblock_kernel_sizes: List[int] = [7, 11],
            source_resblock_dilation_sizes: List[List[int]] = [[1, 3, 5], [1, 3, 5]],
            lrelu_slope: float = 0.1,
            audio_limit: float = 0.99,
            f0_predictor: torch.nn.Module = None,
    ):
        super(HiFTGenerator, self).__init__()

        self.out_channels = 1
        self.nb_harmonics = nb_harmonics
        self.sampling_rate = sampling_rate
        self.istft_params = istft_params
        self.lrelu_slope = lrelu_slope
        self.audio_limit = audio_limit

        self.num_kernels = len(resblock_kernel_sizes)
        self.num_upsamples = len(upsample_rates)
        self.m_source = SourceModuleHnNSF(
            sampling_rate=sampling_rate,
            upsample_scale=np.prod(upsample_rates) * istft_params["hop_len"],
            harmonic_num=nb_harmonics,
            sine_amp=nsf_alpha,
            add_noise_std=nsf_sigma,
            voiced_threshod=nsf_voiced_threshold)
        self.f0_upsamp = torch.nn.Upsample(scale_factor=np.prod(upsample_rates) * istft_params["hop_len"])

        self.conv_pre = weight_norm(
            Conv1d(in_channels, base_channels, 7, 1, padding=3)
        )

        # Up
        self.ups = nn.ModuleList()
        for i, (u, k) in enumerate(zip(upsample_rates, upsample_kernel_sizes)):
            self.ups.append(
                weight_norm(
                    ConvTranspose1d(
                        base_channels // (2**i),
                        base_channels // (2**(i + 1)),
                        k,
                        u,
                        padding=(k - u) // 2,
                    )
                )
            )

        # Down
        self.source_downs = nn.ModuleList()
        self.source_resblocks = nn.ModuleList()
        downsample_rates = [1] + upsample_rates[::-1][:-1]
        downsample_cum_rates = np.cumprod(downsample_rates)
        for i, (u, k, d) in enumerate(zip(downsample_cum_rates[::-1], source_resblock_kernel_sizes, source_resblock_dilation_sizes)):
            if u == 1:
                self.source_downs.append(
                    Conv1d(istft_params["n_fft"] + 2, base_channels // (2 ** (i + 1)), 1, 1)
                )
            else:
                self.source_downs.append(
                    Conv1d(istft_params["n_fft"] + 2, base_channels // (2 ** (i + 1)), u * 2, u, padding=(u // 2))
                )

            self.source_resblocks.append(
                ResBlock(base_channels // (2 ** (i + 1)), k, d)
            )

        self.resblocks = nn.ModuleList()
        for i in range(len(self.ups)):
            ch = base_channels // (2**(i + 1))
            for _, (k, d) in enumerate(zip(resblock_kernel_sizes, resblock_dilation_sizes)):
                self.resblocks.append(ResBlock(ch, k, d))

        self.conv_post = weight_norm(Conv1d(ch, istft_params["n_fft"] + 2, 7, 1, padding=3))
        self.ups.apply(init_weights)
        self.conv_post.apply(init_weights)
        self.reflection_pad = nn.ReflectionPad1d((1, 0))
        self.stft_window = torch.from_numpy(get_window("hann", istft_params["n_fft"], fftbins=True).astype(np.float32))
        self.f0_predictor = f0_predictor

    def remove_weight_norm(self):
        print('Removing weight norm...')
        for l in self.ups:
            remove_weight_norm(l)
        for l in self.resblocks:
            l.remove_weight_norm()
        remove_weight_norm(self.conv_pre)
        remove_weight_norm(self.conv_post)
        self.m_source.remove_weight_norm()
        for l in self.source_downs:
            remove_weight_norm(l)
        for l in self.source_resblocks:
            l.remove_weight_norm()

    def _stft(self, x):
        spec = torch.stft(
            x,
            self.istft_params["n_fft"], self.istft_params["hop_len"], self.istft_params["n_fft"], window=self.stft_window.to(x.device),
            return_complex=True)
        spec = torch.view_as_real(spec)  # [B, F, TT, 2]
        return spec[..., 0], spec[..., 1]

    def _istft(self, magnitude, phase):
        magnitude = torch.clip(magnitude, max=1e2)
        real = magnitude * torch.cos(phase)
        img = magnitude * torch.sin(phase)
        inverse_transform = torch.istft(torch.complex(real, img), self.istft_params["n_fft"], self.istft_params["hop_len"],
                                        self.istft_params["n_fft"], window=self.stft_window.to(magnitude.device))
        return inverse_transform

    def decode(self, x: torch.Tensor, s: torch.Tensor = torch.zeros(1, 1, 0)) -> torch.Tensor:
        s_stft_real, s_stft_imag = self._stft(s.squeeze(1))
        s_stft = torch.cat([s_stft_real, s_stft_imag], dim=1)

        x = self.conv_pre(x)
        for i in range(self.num_upsamples):
            x = F.leaky_relu(x, self.lrelu_slope)
            x = self.ups[i](x)

            if i == self.num_upsamples - 1:
                x = self.reflection_pad(x)

            # fusion
            si = self.source_downs[i](s_stft)
            si = self.source_resblocks[i](si)
            x = x + si

            xs = None
            for j in range(self.num_kernels):
                if xs is None:
                    xs = self.resblocks[i * self.num_kernels + j](x)
                else:
                    xs += self.resblocks[i * self.num_kernels + j](x)
            x = xs / self.num_kernels

        x = F.leaky_relu(x)
        x = self.conv_post(x)
        magnitude = torch.exp(x[:, :self.istft_params["n_fft"] // 2 + 1, :])
        phase = torch.sin(x[:, self.istft_params["n_fft"] // 2 + 1:, :])  # actually, sin is redundancy

        x = self._istft(magnitude, phase)
        x = torch.clamp(x, -self.audio_limit, self.audio_limit)
        return x

    def forward(
            self,
            batch: dict,
            device: torch.device,
    ) -> Dict[str, Optional[torch.Tensor]]:
        speech_feat = batch['speech_feat'].transpose(1, 2).to(device)
        # mel->f0
        f0 = self.f0_predictor(speech_feat)
        # f0->source
        s = self.f0_upsamp(f0[:, None]).transpose(1, 2)  # bs,n,t
        s, _, _ = self.m_source(s)
        s = s.transpose(1, 2)
        # mel+source->speech
        generated_speech = self.decode(x=speech_feat, s=s)
        return generated_speech, f0

    @torch.inference_mode()
    def inference(self, speech_feat: torch.Tensor, cache_source: torch.Tensor = torch.zeros(1, 1, 0)) -> torch.Tensor:
        # mel->f0
        f0 = self.f0_predictor(speech_feat)
        # f0->source
        s = self.f0_upsamp(f0[:, None]).transpose(1, 2)  # bs,n,t
        s, _, _ = self.m_source(s)
        s = s.transpose(1, 2)
        # use cache_source to avoid glitch
        if cache_source.shape[2] != 0:
            s[:, :, :cache_source.shape[2]] = cache_source
        generated_speech = self.decode(x=speech_feat, s=s)
        return generated_speech, s



import unittest

from langchain.prompts import ChatPromptTemplate
from langchain_core.language_models import FakeStreamingListLLM
from apipeline.pipeline.pipeline import Pipeline
from apipeline.pipeline.runner import PipelineRunner
from apipeline.pipeline.task import PipelineTask, PipelineParams
from apipeline.processors.frame_processor import FrameProcessor
from apipeline.frames.sys_frames import StopTaskFrame

from src.types.frames.control_frames import (
    LLMFullResponseEndFrame,
    LLMFullResponseStartFrame,
    UserStartedSpeakingFrame,
    UserStoppedSpeakingFrame,
)
from src.types.frames.data_frames import TextFrame, TranscriptionFrame
from src.processors.aggregators.llm_response import (
    LLMAssistantResponseAggregator,
    LLMUserResponseAggregator,
)
from src.processors.ai_frameworks.langchain_processor import LangchainProcessor


r"""
python -m unittest test.integration.processors.test_langchain_processor.TestLangchainProcessor
"""


class MockProcessor(FrameProcessor):
    def __init__(self, name):
        super().__init__(name=name)
        self.token: list[str] = []
        # Start collecting tokens when we see the start frame
        self.start_collecting = False

    async def process_frame(self, frame, direction):
        await super().process_frame(frame, direction)

        if isinstance(frame, LLMFullResponseStartFrame):
            self.start_collecting = True
        elif isinstance(frame, TextFrame) and self.start_collecting:
            self.token.append(frame.text)
        elif isinstance(frame, LLMFullResponseEndFrame):
            self.start_collecting = False

        await self.push_frame(frame, direction)


class TestLangchainProcessor(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.expected_response = "Hello dear human"
        self.fake_llm = FakeStreamingListLLM(responses=[self.expected_response])

    async def asyncTearDown(self):
        pass

    async def test_langchain(self):
        messages = [("system", "Say hello to {name}"), ("human", "{input}")]
        prompt = ChatPromptTemplate.from_messages(messages).partial(name="Thomas")
        chain = prompt | self.fake_llm
        proc = LangchainProcessor(chain=chain)

        tma_in = LLMUserResponseAggregator(messages)
        tma_out = LLMAssistantResponseAggregator(messages)
        mock_proc = MockProcessor("token_collector")

        pipeline = Pipeline(
            [
                tma_in,
                proc,
                mock_proc,
                tma_out,
            ]
        )

        task = PipelineTask(pipeline, PipelineParams(allow_interruptions=False))
        await task.queue_frames(
            [
                UserStartedSpeakingFrame(),
                TranscriptionFrame(text="Hi World", user_id="user", timestamp="now"),
                UserStoppedSpeakingFrame(),
                StopTaskFrame(),
            ]
        )

        runner = PipelineRunner()
        await runner.run(task)
        self.assertEqual("".join(mock_proc.token), self.expected_response)



from functools import reduce
import math
import numpy as np
import torch
from torch import nn
from torch.nn import functional as F

from torch.backends.cuda import sdp_kernel
from packaging import version

from dac.nn.layers import Snake1d

class ResidualBlock(nn.Module):
    def __init__(self, main, skip=None):
        super().__init__()
        self.main = nn.Sequential(*main)
        self.skip = skip if skip else nn.Identity()

    def forward(self, input):
        return self.main(input) + self.skip(input)

class ResConvBlock(ResidualBlock):
    def __init__(self, c_in, c_mid, c_out, is_last=False, kernel_size=5, conv_bias=True, use_snake=False):
        skip = None if c_in == c_out else nn.Conv1d(c_in, c_out, 1, bias=False)
        super().__init__([
            nn.Conv1d(c_in, c_mid, kernel_size, padding=kernel_size//2, bias=conv_bias),
            nn.GroupNorm(1, c_mid),
            Snake1d(c_mid) if use_snake else nn.GELU(),
            nn.Conv1d(c_mid, c_out, kernel_size, padding=kernel_size//2, bias=conv_bias),
            nn.GroupNorm(1, c_out) if not is_last else nn.Identity(),
            (Snake1d(c_out) if use_snake else nn.GELU()) if not is_last else nn.Identity(),
        ], skip)

class SelfAttention1d(nn.Module):
    def __init__(self, c_in, n_head=1, dropout_rate=0.):
        super().__init__()
        assert c_in % n_head == 0
        self.norm = nn.GroupNorm(1, c_in)
        self.n_head = n_head
        self.qkv_proj = nn.Conv1d(c_in, c_in * 3, 1)
        self.out_proj = nn.Conv1d(c_in, c_in, 1)
        self.dropout = nn.Dropout(dropout_rate, inplace=True)

        self.use_flash = torch.cuda.is_available() and version.parse(torch.__version__) >= version.parse('2.0.0')

        if not self.use_flash:
            return

        device_properties = torch.cuda.get_device_properties(torch.device('cuda'))

        if device_properties.major == 8 and device_properties.minor == 0:
            # Use flash attention for A100 GPUs
            self.sdp_kernel_config = (True, False, False)
        else:
            # Don't use flash attention for other GPUs
            self.sdp_kernel_config = (False, True, True)

    def forward(self, input):
        n, c, s = input.shape
        qkv = self.qkv_proj(self.norm(input))
        qkv = qkv.view(
            [n, self.n_head * 3, c // self.n_head, s]).transpose(2, 3)
        q, k, v = qkv.chunk(3, dim=1)
        scale = k.shape[3]**-0.25

        if self.use_flash:
            with sdp_kernel(*self.sdp_kernel_config):
                y = F.scaled_dot_product_attention(q, k, v, is_causal=False).contiguous().view([n, c, s])
        else:
            att = ((q * scale) @ (k.transpose(2, 3) * scale)).softmax(3)
            y = (att @ v).transpose(2, 3).contiguous().view([n, c, s])


        return input + self.dropout(self.out_proj(y))

class SkipBlock(nn.Module):
    def __init__(self, *main):
        super().__init__()
        self.main = nn.Sequential(*main)

    def forward(self, input):
        return torch.cat([self.main(input), input], dim=1)

class FourierFeatures(nn.Module):
    def __init__(self, in_features, out_features, std=1.):
        super().__init__()
        assert out_features % 2 == 0
        self.weight = nn.Parameter(torch.randn(
            [out_features // 2, in_features]) * std)

    def forward(self, input):
        f = 2 * math.pi * input @ self.weight.T
        return torch.cat([f.cos(), f.sin()], dim=-1)

def expand_to_planes(input, shape):
    return input[..., None].repeat([1, 1, shape[2]])

_kernels = {
    'linear':
        [1 / 8, 3 / 8, 3 / 8, 1 / 8],
    'cubic': 
        [-0.01171875, -0.03515625, 0.11328125, 0.43359375,
        0.43359375, 0.11328125, -0.03515625, -0.01171875],
    'lanczos3': 
        [0.003689131001010537, 0.015056144446134567, -0.03399861603975296,
        -0.066637322306633, 0.13550527393817902, 0.44638532400131226,
        0.44638532400131226, 0.13550527393817902, -0.066637322306633,
        -0.03399861603975296, 0.015056144446134567, 0.003689131001010537]
}

class Downsample1d(nn.Module):
    def __init__(self, kernel='linear', pad_mode='reflect', channels_last=False):
        super().__init__()
        self.pad_mode = pad_mode
        kernel_1d = torch.tensor(_kernels[kernel])
        self.pad = kernel_1d.shape[0] // 2 - 1
        self.register_buffer('kernel', kernel_1d)
        self.channels_last = channels_last
    
    def forward(self, x):
        if self.channels_last:
            x = x.permute(0, 2, 1)
        x = F.pad(x, (self.pad,) * 2, self.pad_mode)
        weight = x.new_zeros([x.shape[1], x.shape[1], self.kernel.shape[0]])
        indices = torch.arange(x.shape[1], device=x.device)
        weight[indices, indices] = self.kernel.to(weight)
        x = F.conv1d(x, weight, stride=2)
        if self.channels_last:
            x = x.permute(0, 2, 1)
        return x


class Upsample1d(nn.Module):
    def __init__(self, kernel='linear', pad_mode='reflect', channels_last=False):
        super().__init__()
        self.pad_mode = pad_mode
        kernel_1d = torch.tensor(_kernels[kernel]) * 2
        self.pad = kernel_1d.shape[0] // 2 - 1
        self.register_buffer('kernel', kernel_1d)
        self.channels_last = channels_last
    
    def forward(self, x):
        if self.channels_last:
            x = x.permute(0, 2, 1)
        x = F.pad(x, ((self.pad + 1) // 2,) * 2, self.pad_mode)
        weight = x.new_zeros([x.shape[1], x.shape[1], self.kernel.shape[0]])
        indices = torch.arange(x.shape[1], device=x.device)
        weight[indices, indices] = self.kernel.to(weight)
        x = F.conv_transpose1d(x, weight, stride=2, padding=self.pad * 2 + 1)
        if self.channels_last:
            x = x.permute(0, 2, 1)
        return x
        
def Downsample1d_2(
    in_channels: int, out_channels: int, factor: int, kernel_multiplier: int = 2
) -> nn.Module:
    assert kernel_multiplier % 2 == 0, "Kernel multiplier must be even"

    return nn.Conv1d(
        in_channels=in_channels,
        out_channels=out_channels,
        kernel_size=factor * kernel_multiplier + 1,
        stride=factor,
        padding=factor * (kernel_multiplier // 2),
    )


def Upsample1d_2(
    in_channels: int, out_channels: int, factor: int, use_nearest: bool = False
) -> nn.Module:

    if factor == 1:
        return nn.Conv1d(
            in_channels=in_channels, out_channels=out_channels, kernel_size=3, padding=1
        )

    if use_nearest:
        return nn.Sequential(
            nn.Upsample(scale_factor=factor, mode="nearest"),
            nn.Conv1d(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=3,
                padding=1,
            ),
        )
    else:
        return nn.ConvTranspose1d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=factor * 2,
            stride=factor,
            padding=factor // 2 + factor % 2,
            output_padding=factor % 2,
        )

def zero_init(layer):
    nn.init.zeros_(layer.weight)
    if layer.bias is not None:
        nn.init.zeros_(layer.bias)
    return layer

def rms_norm(x, scale, eps):
    dtype = reduce(torch.promote_types, (x.dtype, scale.dtype, torch.float32))
    mean_sq = torch.mean(x.to(dtype)**2, dim=-1, keepdim=True)
    scale = scale.to(dtype) * torch.rsqrt(mean_sq + eps)
    return x * scale.to(x.dtype)

#rms_norm = torch.compile(rms_norm)

class AdaRMSNorm(nn.Module):
    def __init__(self, features, cond_features, eps=1e-6):
        super().__init__()
        self.eps = eps
        self.linear = zero_init(nn.Linear(cond_features, features, bias=False))
  
    def extra_repr(self):
        return f"eps={self.eps},"

    def forward(self, x, cond):
        return rms_norm(x, self.linear(cond)[:, None, :] + 1, self.eps)
    
def normalize(x, eps=1e-4):
    dim = list(range(1, x.ndim))
    n = torch.linalg.vector_norm(x, dim=dim, keepdim=True)
    alpha = np.sqrt(n.numel() / x.numel())
    return x / torch.add(eps, n, alpha=alpha)

class ForcedWNConv1d(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size=1):
        super().__init__()
        self.weight = nn.Parameter(torch.randn([out_channels, in_channels, kernel_size]))

    def forward(self, x):
        if self.training:
            with torch.no_grad():
                self.weight.copy_(normalize(self.weight))
        
        fan_in = self.weight[0].numel()

        w = normalize(self.weight) / math.sqrt(fan_in)

        return F.conv1d(x, w, padding='same')
        
# Kernels

use_compile = True

def compile(function, *args, **kwargs):
    if not use_compile:
        return function
    try:
        return torch.compile(function, *args, **kwargs)
    except RuntimeError:
        return function


@compile
def linear_geglu(x, weight, bias=None):
    x = x @ weight.mT
    if bias is not None:
        x = x + bias
    x, gate = x.chunk(2, dim=-1)
    return x * F.gelu(gate)


@compile
def rms_norm(x, scale, eps):
    dtype = reduce(torch.promote_types, (x.dtype, scale.dtype, torch.float32))
    mean_sq = torch.mean(x.to(dtype)**2, dim=-1, keepdim=True)
    scale = scale.to(dtype) * torch.rsqrt(mean_sq + eps)
    return x * scale.to(x.dtype)

# Layers

class LinearGEGLU(nn.Linear):
    def __init__(self, in_features, out_features, bias=True):
        super().__init__(in_features, out_features * 2, bias=bias)
        self.out_features = out_features

    def forward(self, x):
        return linear_geglu(x, self.weight, self.bias)


class RMSNorm(nn.Module):
    def __init__(self, shape, fix_scale = False, eps=1e-6):
        super().__init__()
        self.eps = eps

        if fix_scale:
            self.register_buffer("scale", torch.ones(shape))
        else:
            self.scale = nn.Parameter(torch.ones(shape))

    def extra_repr(self):
        return f"shape={tuple(self.scale.shape)}, eps={self.eps}"

    def forward(self, x):
        return rms_norm(x, self.scale, self.eps)    

def snake_beta(x, alpha, beta):
    return x + (1.0 / (beta + 0.000000001)) * pow(torch.sin(x * alpha), 2)

# try:
#     snake_beta = torch.compile(snake_beta)
# except RuntimeError:
#     pass

# Adapted from https://github.com/NVIDIA/BigVGAN/blob/main/activations.py under MIT license
# License available in LICENSES/LICENSE_NVIDIA.txt
class SnakeBeta(nn.Module):

    def __init__(self, in_features, alpha=1.0, alpha_trainable=True, alpha_logscale=True):
        super(SnakeBeta, self).__init__()
        self.in_features = in_features

        # initialize alpha
        self.alpha_logscale = alpha_logscale
        if self.alpha_logscale: # log scale alphas initialized to zeros
            self.alpha = nn.Parameter(torch.zeros(in_features) * alpha)
            self.beta = nn.Parameter(torch.zeros(in_features) * alpha)
        else: # linear scale alphas initialized to ones
            self.alpha = nn.Parameter(torch.ones(in_features) * alpha)
            self.beta = nn.Parameter(torch.ones(in_features) * alpha)

        self.alpha.requires_grad = alpha_trainable
        self.beta.requires_grad = alpha_trainable

        self.no_div_by_zero = 0.000000001

    def forward(self, x):
        alpha = self.alpha.unsqueeze(0).unsqueeze(-1) # line up with x to [B, C, T]
        beta = self.beta.unsqueeze(0).unsqueeze(-1)
        if self.alpha_logscale:
            alpha = torch.exp(alpha)
            beta = torch.exp(beta)
        x = snake_beta(x, alpha, beta)

        return x


r"""
mkdir -p records && python -m demo.tts_chat
mkdir -p records && STEAM=1 python -m demo.tts_chat
"""

import logging
import os

import torch
import torchaudio
import numpy as np

import deps.ChatTTS.ChatTTS as ChatTTS
from deps.ChatTTS.ChatTTS.core import Chat
from deps.ChatTTS.tools.normalizer.en import normalizer_en_nemo_text
from deps.ChatTTS.tools.normalizer.zh import normalizer_zh_tn


def load_normalizer(chat: Chat):
    # try to load normalizer
    try:
        chat.normalizer.register("en", normalizer_en_nemo_text())
    except ValueError as e:
        logging.error(e)
    except BaseException:
        logging.warning("Package nemo_text_processing not found!")
        logging.warning(
            "Run: conda install -c conda-forge pynini=2.1.5 && pip install nemo_text_processing",
        )
    try:
        chat.normalizer.register("zh", normalizer_zh_tn())
    except ValueError as e:
        logging.error(e)
    except BaseException:
        logging.warning("Package WeTextProcessing not found!")
        logging.warning(
            "Run: conda install -c conda-forge pynini=2.1.5 && pip install WeTextProcessing",
        )


if __name__ == "__main__":
    is_stream = bool(os.getenv("STREAM", ""))

    chat = ChatTTS.Chat()
    load_normalizer(chat)

    chat.load(compile=True, source="custom", custom_path="./models/2Noise/ChatTTS")

    texts = ["你好，我是机器人", "我是机器人"]
    ###################################
    # Sample a speaker from Gaussian.
    rand_spk = chat.sample_random_speaker()
    params_infer_code = Chat.InferCodeParams(
        spk_emb=rand_spk,  # add sampled speaker
        top_P=0.7,  # top P decode
        top_K=20,  # top K decode
        temperature=0.3,  # using custom temperature
    )

    ###################################
    # For sentence level manual control.

    # use oral_(0-9), laugh_(0-2), break_(0-7)
    # to generate special token in text to synthesize.
    params_refine_text = Chat.RefineTextParams(
        prompt="[oral_2][laugh_0][break_6]",
    )

    if is_stream is False:
        wavs = chat.infer(
            texts,
            params_refine_text=params_refine_text,
            params_infer_code=params_infer_code,
            do_text_normalization=False,
        )
        print("wavs[0] type", type(wavs[0]), "wavs[0] shape", wavs[0].shape, "shape", wavs.shape)

        # Assuming 'wavs' is a NumPy array of shape (num_samples,) or (num_channels, num_samples)
        wav_tensor = torch.from_numpy(wavs[0].astype(np.float32))  # Ensure data is float32

        # If 'wavs' is 1D, add a channel dimension
        if wav_tensor.dim() == 1:
            wav_tensor = wav_tensor.unsqueeze(0)
        print("wav tensor shape", wav_tensor.shape, "type", wav_tensor.dtype)

        torchaudio.save("./records/chat_tts_output1.wav", wav_tensor, 24000)

        ###################################
        # For word level manual control.
        text = "What is [uv_break]your favorite english food.[laugh][lbreak]"
        wav = chat.infer(
            text,
            skip_refine_text=True,
            params_refine_text=params_refine_text,
            params_infer_code=params_infer_code,
            do_text_normalization=False,
        )
        print("wav[0] type", type(wav[0]), "shape", wav.shape)

        # Assuming 'wavs' is a NumPy array of shape (num_samples,) or (num_channels, num_samples)
        wav_tensor = torch.from_numpy(wav[0].astype(np.float32))  # Ensure data is float32

        # If 'wavs' is 1D, add a channel dimension
        if wav_tensor.dim() == 1:
            wav_tensor = wav_tensor.unsqueeze(0)
        print("wav tensor shape", wav_tensor.shape, "type", wav_tensor.dtype)

        torchaudio.save("./records/chat_tts_output2.wav", wav_tensor, 24000)

    else:
        ############ stream infer ############
        texts = ["你好，我是机器人", "我是机器人一号", "我是机器人二号"]
        wavs_iter = chat.infer(
            texts,
            stream=True,
            params_refine_text=params_refine_text,
            params_infer_code=params_infer_code,
            do_text_normalization=False,
        )
        res = None
        for wavs in wavs_iter:
            if res is None:
                res = wavs
            else:
                res = np.concatenate([res, wavs], axis=1)

            print("wavs type", type(wavs), "shape", wavs.shape)

        print("res type", type(res), "shape", res.shape)

        for i in range(res.shape[0]):
            # Assuming 'wavs' is a NumPy array of shape (num_samples,) or
            # (num_channels, num_samples)
            wav_tensor = torch.from_numpy(res[i].astype(np.float32))  # Ensure data is float32

            # If 'wavs' is 1D, add a channel dimension
            if wav_tensor.dim() == 1:
                wav_tensor = wav_tensor.unsqueeze(0)
            print("wav tensor shape", wav_tensor.shape, "type", wav_tensor.dtype)

            torchaudio.save(f"./records/chat_tts_stream{i}.wav", wav_tensor, 24000)



import argparse
from transformers import AutoTokenizer, AutoModelForCausalLM
import torch
import os

from tqdm import tqdm
from PIL import Image
import json
import os
import requests
from PIL import Image
from io import BytesIO
import math

import argparse
from transformers import AutoTokenizer, AutoModelForCausalLM
import torch
import os
from GOT.utils.conversation import conv_templates, SeparatorStyle
from GOT.utils.utils import disable_torch_init
from transformers import CLIPVisionModel, CLIPImageProcessor, StoppingCriteria
from GOT.model import *
from GOT.utils.utils import KeywordsStoppingCriteria

from PIL import Image

import os
import requests
from PIL import Image
from io import BytesIO
from GOT.model.plug.blip_process import BlipImageEvalProcessor

from transformers import TextStreamer
from GOT.model.plug.transforms import train_transform, test_transform
import re
from GOT.demo.process_results import punctuation_dict, svg_to_html

DEFAULT_IMAGE_TOKEN = "<image>"
DEFAULT_IMAGE_PATCH_TOKEN = '<imgpad>'
DEFAULT_IM_START_TOKEN = '<img>'
DEFAULT_IM_END_TOKEN = '</img>'


import string
 
translation_table = str.maketrans(punctuation_dict)


def load_image(image_file):
    if image_file.startswith('http') or image_file.startswith('https'):
        response = requests.get(image_file)
        image = Image.open(BytesIO(response.content)).convert('RGB')
    else:
        image = Image.open(image_file).convert('RGB')
    return image


def find_closest_aspect_ratio(aspect_ratio, target_ratios, width, height, image_size):
    best_ratio_diff = float('inf')
    best_ratio = (1, 1)
    area = width * height
    for ratio in target_ratios:
        target_aspect_ratio = ratio[0] / ratio[1]
        ratio_diff = abs(aspect_ratio - target_aspect_ratio)
        if ratio_diff < best_ratio_diff:
            best_ratio_diff = ratio_diff
            best_ratio = ratio
        elif ratio_diff == best_ratio_diff:
            if area > 0.5 * image_size * image_size * ratio[0] * ratio[1]:
                best_ratio = ratio
    # print(f'width: {width}, height: {height}, best_ratio: {best_ratio}')
    return best_ratio


def dynamic_preprocess(image, min_num=1, max_num=6, image_size=1024, use_thumbnail=True):
    orig_width, orig_height = image.size
    aspect_ratio = orig_width / orig_height

    # calculate the existing image aspect ratio
    target_ratios = set(
        (i, j) for n in range(min_num, max_num + 1) for i in range(1, n + 1) for j in range(1, n + 1) if
        i * j <= max_num and i * j >= min_num)
    # print(target_ratios)
    target_ratios = sorted(target_ratios, key=lambda x: x[0] * x[1])

    # find the closest aspect ratio to the target
    target_aspect_ratio = find_closest_aspect_ratio(
        aspect_ratio, target_ratios, orig_width, orig_height, image_size)

    # print(target_aspect_ratio)
    # calculate the target width and height
    target_width = image_size * target_aspect_ratio[0]
    target_height = image_size * target_aspect_ratio[1]
    blocks = target_aspect_ratio[0] * target_aspect_ratio[1]

    # print(blocks)

    # resize the image
    resized_img = image.resize((target_width, target_height))
    processed_images = []
    for i in range(blocks):
        box = (
            (i % (target_width // image_size)) * image_size,
            (i // (target_width // image_size)) * image_size,
            ((i % (target_width // image_size)) + 1) * image_size,
            ((i // (target_width // image_size)) + 1) * image_size
        )
        # split the image
        split_img = resized_img.crop(box)
        processed_images.append(split_img)
    assert len(processed_images) == blocks
    if use_thumbnail and len(processed_images) != 1:
        thumbnail_img = image.resize((image_size, image_size))
        processed_images.append(thumbnail_img)
    return processed_images



def split_list(lst, n):
    """Split a list into n (roughly) equal-sized chunks"""
    chunk_size = math.ceil(len(lst) / n)  # integer division
    return [lst[i:i+chunk_size] for i in range(0, len(lst), chunk_size)]


def get_chunk(lst, n, k):
    chunks = split_list(lst, n)
    return chunks[k]



output_list = []

def eval_model(args):
    # Model
    disable_torch_init()
    model_name = os.path.expanduser(args.model_name)
    tokenizer = AutoTokenizer.from_pretrained(model_name, trust_remote_code=True)


    model = GOTQwenForCausalLM.from_pretrained(model_name, low_cpu_mem_usage=True, device_map='cuda', use_safetensors=True, pad_token_id=151643).eval()


    # vary old codes, no use
    image_processor = BlipImageEvalProcessor(image_size=1024)


    # image_processor_high = BlipImageEvalProcessor(image_size=1280)
    image_processor_high = BlipImageEvalProcessor(image_size=1024)
    use_im_start_end = True



    # image_token_len = 400
    image_token_len = 256
    gts_path = args.gtfile_path
    gts = json.load(open(gts_path))

    # gts = gts[0]


    print("Generate Results......")


    if "OCR" in args.datatype:
            gts = get_chunk(gts, args.num_chunks, args.chunk_idx)


    for ann in tqdm(gts):
        output_json = {}
        
        if "OCR" in args.datatype:
            qs = ann["conversations"][0]["value"]
        else:
            qs = ann["question"]
            # ans = ann["answers"][0]
        
        qs2 = qs
        image_file = ann["image"] 
        if 'Text' in args.datatype:
            image_file = image_file + '.jpg'
        if "VQAv2" in args.datatype:
            image_file = 'COCO_' + 'val2014' + '_'+ str(image_file).zfill(12) + '.jpg'
        if "Cap" in args.datatype:
            image_file = 'COCO_' + 'val2014' + '_'+ str(image_file).zfill(12) + '.jpg'

        image_file_path = os.path.join(args.image_path, image_file)
        # print(image_file_path)
        # exit()

        # qs = args.query
        # if mm_use_im_start_end:



        multi_crop = False
        if multi_crop:
            image_list = []
            # qs =  DEFAULT_IM_START_TOKEN + DEFAULT_IMAGE_PATCH_TOKEN * image_token_len + DEFAULT_IM_END_TOKEN  + '\n' +  'OCR with format upon the patch reference: '
            img = load_image(image_file_path)
            sub_images = dynamic_preprocess(img)
            ll = len(sub_images)
            for p in sub_images:
                image = p
                image_1 = image.copy()
                # vary old code, NO USE
                image_tensor = image_processor_high(image_1)

                # image_tensor_1 = image_processor_high.preprocess(image_1, return_tensors='pt')['pixel_values'][0]

                image_tensor_1 = image_processor_high(image_1)

                image_list.append(image_tensor_1)

                # print(image_tensor_1.shape)

            image_list = torch.stack(image_list)

        else:
            ll = 1
            image = load_image(image_file_path)
            image_1 = image.copy()
            # image_1 = image_1.resize((1024, 1024))

            # vary old code, NO USE
            image_tensor = image_processor_high(image_1)

            image_tensor_1 = image_processor_high(image_1)
            # image_tensor_1 = torch.zeros(3, 1024, 1024)


        qs =  DEFAULT_IM_START_TOKEN + DEFAULT_IMAGE_PATCH_TOKEN * image_token_len*ll + DEFAULT_IM_END_TOKEN  + '\n' +  'OCR with format: '

        

        conv_mode = "mpt"

        if args.conv_mode is not None and conv_mode != args.conv_mode:
            print('[WARNING] the auto inferred conversation mode is {}, while `--conv-mode` is {}, using {}'.format(conv_mode, args.conv_mode, args.conv_mode))
        else:
            args.conv_mode = conv_mode

        conv = conv_templates[args.conv_mode].copy()
        conv.append_message(conv.roles[0], qs)
        conv.append_message(conv.roles[1], None)
        prompt = conv.get_prompt()
        inputs = tokenizer([prompt])

        input_ids = torch.as_tensor(inputs.input_ids).cuda()

        stop_str = conv.sep if conv.sep_style != SeparatorStyle.TWO else conv.sep2
        keywords = [stop_str]
        stopping_criteria = KeywordsStoppingCriteria(keywords, tokenizer, input_ids)

        if multi_crop:
            with torch.autocast("cuda", dtype=torch.bfloat16):
                output_ids = model.generate(
                    input_ids,
                    images=[(image_list.half().cuda(), image_list.half().cuda())],
                    do_sample=False,
                    num_beams = 1,
                    # temperature=0.2,
                    # no_repeat_ngram_size = 20,
                    # streamer=streamer,
                    max_new_tokens=4096,
                    stopping_criteria=[stopping_criteria]
                    )
        else:
            with torch.autocast("cuda", dtype=torch.bfloat16):
                output_ids = model.generate(
                    input_ids,
                    images=[(image_tensor.unsqueeze(0).half().cuda(), image_tensor_1.unsqueeze(0).half().cuda())],
                    do_sample=False,
                    num_beams = 1,
                    # temperature=0.2,
                    no_repeat_ngram_size = 20,
                    # encoder_repetition_penalty = 1.2,
                    # penalty_alpha=0.2,
                    # top_k=3,
                    max_new_tokens=4096,
                    stopping_criteria=[stopping_criteria]
                    )

        outputs = tokenizer.decode(output_ids[0, input_ids.shape[1]:]).strip()
        
        if outputs.endswith(stop_str):
            outputs = outputs[:-len(stop_str)]
        outputs = outputs.strip()
        # outputs = outputs.strip()[:-1]
        if "Cap" in args.datatype:
            # output_json['image'] = ann["image"]
            output_json['image_id'] = ann["id"]
            output_json["caption"] = outputs
        else:
            # output_json['questionId'] = qs_id
            # output_json['question_id'] = qs_id
            output_json['image'] = ann["image"]
            output_json['question'] = qs 
            output_json['label'] = ann["conversations"][1]["value"]
            output_json['answer'] = outputs
        output_list.append(output_json)

    filename = args.out_path + "/results_" + str(args.chunk_idx) + ".json"
    with open(filename, 'w', encoding="utf-8") as file_obj:
        json.dump(output_list, file_obj, ensure_ascii=False, indent=1)
        # print(outputs)
    # print("Evaluate Results... ")
    # doc_text_eval(gts_path, filename, args.datatype)

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--model-name", type=str, default="facebook/opt-350m")
    parser.add_argument("--gtfile_path", type=str, required=True)
    parser.add_argument("--image_path", type=str, required=True)
    parser.add_argument("--out_path", type=str, required=True)
    parser.add_argument("--datatype", type=str, required=True)  # Text or Doc
    parser.add_argument("--num-chunks", type=int, default=1)
    parser.add_argument("--chunk-idx", type=int, default=0)
    # parser.add_argument("--query", type=str, required=True)
    parser.add_argument("--conv-mode", type=str, default=None)
    parser.add_argument("--temperature", type=float, default=0.2)
    args = parser.parse_args()
    print(args)
    eval_model(args)



import math
from pathlib import Path
from random import Random

import click
from loguru import logger
from pydub import AudioSegment
from tqdm import tqdm

from fish_speech.utils.file import AUDIO_EXTENSIONS, list_files, load_filelist


@click.command()
@click.argument("root", type=click.Path(exists=True, path_type=Path))
@click.option("--val-ratio", type=float, default=None)
@click.option("--val-count", type=int, default=None)
@click.option("--filelist", default=None, type=Path)
@click.option("--min-duration", default=None, type=float)
@click.option("--max-duration", default=None, type=float)
def main(root, val_ratio, val_count, filelist, min_duration, max_duration):
    if filelist:
        files = [i[0] for i in load_filelist(filelist)]
    else:
        files = list_files(root, AUDIO_EXTENSIONS, recursive=True, sort=True)

    if min_duration is None and max_duration is None:
        filtered_files = list(map(str, [file.relative_to(root) for file in files]))
    else:
        filtered_files = []
        for file in tqdm(files):
            try:
                audio = AudioSegment.from_file(str(file))
                duration = len(audio) / 1000.0

                if min_duration is not None and duration < min_duration:
                    logger.info(
                        f"Skipping {file} due to duration {duration:.2f} < {min_duration:.2f}"
                    )
                    continue

                if max_duration is not None and duration > max_duration:
                    logger.info(
                        f"Skipping {file} due to duration {duration:.2f} > {max_duration:.2f}"
                    )
                    continue

                filtered_files.append(str(file.relative_to(root)))
            except Exception as e:
                logger.info(f"Error processing {file}: {e}")

    logger.info(
        f"Found {len(files)} files, remaining {len(filtered_files)} files after filtering"
    )

    Random(42).shuffle(filtered_files)

    if val_count is None and val_ratio is None:
        logger.info("Validation ratio and count not specified, using min(20%, 100)")
        val_size = min(100, math.ceil(len(filtered_files) * 0.2))
    elif val_count is not None and val_ratio is not None:
        logger.error("Cannot specify both val_count and val_ratio")
        return
    elif val_count is not None:
        if val_count < 1 or val_count > len(filtered_files):
            logger.error("val_count must be between 1 and number of files")
            return
        val_size = val_count
    else:
        val_size = math.ceil(len(filtered_files) * val_ratio)

    logger.info(f"Using {val_size} files for validation")

    with open(root / "vq_train_filelist.txt", "w", encoding="utf-8") as f:
        f.write("\n".join(filtered_files[val_size:]))

    with open(root / "vq_val_filelist.txt", "w", encoding="utf-8") as f:
        f.write("\n".join(filtered_files[:val_size]))

    logger.info("Done")


if __name__ == "__main__":
    main()



import base64
import os
from functools import lru_cache
from typing import Optional
import torch
from transformers import AutoTokenizer
from whisper.tokenizer import Tokenizer

import tiktoken

LANGUAGES = {
    "en": "english",
    "zh": "chinese",
    "de": "german",
    "es": "spanish",
    "ru": "russian",
    "ko": "korean",
    "fr": "french",
    "ja": "japanese",
    "pt": "portuguese",
    "tr": "turkish",
    "pl": "polish",
    "ca": "catalan",
    "nl": "dutch",
    "ar": "arabic",
    "sv": "swedish",
    "it": "italian",
    "id": "indonesian",
    "hi": "hindi",
    "fi": "finnish",
    "vi": "vietnamese",
    "he": "hebrew",
    "uk": "ukrainian",
    "el": "greek",
    "ms": "malay",
    "cs": "czech",
    "ro": "romanian",
    "da": "danish",
    "hu": "hungarian",
    "ta": "tamil",
    "no": "norwegian",
    "th": "thai",
    "ur": "urdu",
    "hr": "croatian",
    "bg": "bulgarian",
    "lt": "lithuanian",
    "la": "latin",
    "mi": "maori",
    "ml": "malayalam",
    "cy": "welsh",
    "sk": "slovak",
    "te": "telugu",
    "fa": "persian",
    "lv": "latvian",
    "bn": "bengali",
    "sr": "serbian",
    "az": "azerbaijani",
    "sl": "slovenian",
    "kn": "kannada",
    "et": "estonian",
    "mk": "macedonian",
    "br": "breton",
    "eu": "basque",
    "is": "icelandic",
    "hy": "armenian",
    "ne": "nepali",
    "mn": "mongolian",
    "bs": "bosnian",
    "kk": "kazakh",
    "sq": "albanian",
    "sw": "swahili",
    "gl": "galician",
    "mr": "marathi",
    "pa": "punjabi",
    "si": "sinhala",
    "km": "khmer",
    "sn": "shona",
    "yo": "yoruba",
    "so": "somali",
    "af": "afrikaans",
    "oc": "occitan",
    "ka": "georgian",
    "be": "belarusian",
    "tg": "tajik",
    "sd": "sindhi",
    "gu": "gujarati",
    "am": "amharic",
    "yi": "yiddish",
    "lo": "lao",
    "uz": "uzbek",
    "fo": "faroese",
    "ht": "haitian creole",
    "ps": "pashto",
    "tk": "turkmen",
    "nn": "nynorsk",
    "mt": "maltese",
    "sa": "sanskrit",
    "lb": "luxembourgish",
    "my": "myanmar",
    "bo": "tibetan",
    "tl": "tagalog",
    "mg": "malagasy",
    "as": "assamese",
    "tt": "tatar",
    "haw": "hawaiian",
    "ln": "lingala",
    "ha": "hausa",
    "ba": "bashkir",
    "jw": "javanese",
    "su": "sundanese",
    "yue": "cantonese",
    "minnan": "minnan",
    "wuyu": "wuyu",
    "dialect": "dialect",
    "zh/en": "zh/en",
    "en/zh": "en/zh",
}

# language code lookup by name, with a few language aliases
TO_LANGUAGE_CODE = {
    **{language: code for code, language in LANGUAGES.items()},
    "burmese": "my",
    "valencian": "ca",
    "flemish": "nl",
    "haitian": "ht",
    "letzeburgesch": "lb",
    "pushto": "ps",
    "panjabi": "pa",
    "moldavian": "ro",
    "moldovan": "ro",
    "sinhalese": "si",
    "castilian": "es",
    "mandarin": "zh",
}

AUDIO_EVENT = {
    "ASR": "ASR",
    "AED": "AED",
    "SER": "SER",
    "Speech": "Speech",
    "/Speech": "/Speech",
    "BGM": "BGM",
    "/BGM": "/BGM",
    "Laughter": "Laughter",
    "/Laughter": "/Laughter",
    "Applause": "Applause",
    "/Applause": "/Applause",
}

EMOTION = {
    "HAPPY": "HAPPY",
    "SAD": "SAD",
    "ANGRY": "ANGRY",
    "NEUTRAL": "NEUTRAL",
}

TTS_Vocal_Token = {
    "TTS/B": "TTS/B",
    "TTS/O": "TTS/O",
    "TTS/Q": "TTS/Q",
    "TTS/A": "TTS/A",
    "TTS/CO": "TTS/CO",
    "TTS/CL": "TTS/CL",
    "TTS/H": "TTS/H",
    **{f"TTS/SP{i:02d}": f"TTS/SP{i:02d}" for i in range(1, 14)}
}


@lru_cache(maxsize=None)
def get_encoding(name: str = "gpt2", num_languages: int = 99):
    vocab_path = os.path.join(os.path.dirname(__file__), "assets", f"{name}.tiktoken")
    ranks = {
        base64.b64decode(token): int(rank)
        for token, rank in (line.split() for line in open(vocab_path) if line)
    }
    n_vocab = len(ranks)
    special_tokens = {}

    specials = [
        "<|endoftext|>",
        "<|startoftranscript|>",
        *[f"<|{lang}|>" for lang in list(LANGUAGES.keys())[:num_languages]],
        *[f"<|{audio_event}|>" for audio_event in list(AUDIO_EVENT.keys())],
        *[f"<|{emotion}|>" for emotion in list(EMOTION.keys())],
        "<|translate|>",
        "<|transcribe|>",
        "<|startoflm|>",
        "<|startofprev|>",
        "<|nospeech|>",
        "<|notimestamps|>",
        *[f"<|SPECIAL_TOKEN_{i}|>" for i in range(1, 31)],        # register special tokens for ASR
        *[f"<|{tts}|>" for tts in list(TTS_Vocal_Token.keys())],  # register special tokens for TTS
        *[f"<|{i * 0.02:.2f}|>" for i in range(1501)],
    ]

    for token in specials:
        special_tokens[token] = n_vocab
        n_vocab += 1

    return tiktoken.Encoding(
        name=os.path.basename(vocab_path),
        explicit_n_vocab=n_vocab,
        pat_str=r"""'s|'t|'re|'ve|'m|'ll|'d| ?\p{L}+| ?\p{N}+| ?[^\s\p{L}\p{N}]+|\s+(?!\S)|\s+""",
        mergeable_ranks=ranks,
        special_tokens=special_tokens,
    )


@lru_cache(maxsize=None)
def get_tokenizer(
    multilingual: bool,
    *,
    num_languages: int = 99,
    language: Optional[str] = None,
    task: Optional[str] = None,  # Literal["transcribe", "translate", None]
) -> Tokenizer:
    if language is not None:
        language = language.lower()
        if language not in LANGUAGES:
            if language in TO_LANGUAGE_CODE:
                language = TO_LANGUAGE_CODE[language]
            else:
                raise ValueError(f"Unsupported language: {language}")

    if multilingual:
        encoding_name = "multilingual_zh_ja_yue_char_del"
        language = language or "en"
        task = task or "transcribe"
    else:
        encoding_name = "gpt2"
        language = None
        task = None

    encoding = get_encoding(name=encoding_name, num_languages=num_languages)

    return Tokenizer(
        encoding=encoding, num_languages=num_languages, language=language, task=task
    )


class QwenTokenizer():
    def __init__(self, token_path, skip_special_tokens=True):
        super().__init__()
        # NOTE: non-chat model, all these special tokens keep randomly initialized.
        special_tokens = {
            'eos_token': '<|endoftext|>',
            'pad_token': '<|endoftext|>',
            'additional_special_tokens': [
                '<|im_start|>', '<|im_end|>', '<|endofprompt|>',
                '[breath]', '<strong>', '</strong>', '[noise]',
                '[laughter]', '[cough]', '[clucking]', '[accent]',
                '[quick_breath]',
                "<laughter>", "</laughter>",
                "[hissing]", "[sigh]", "[vocalized-noise]",
                "[lipsmack]", "[mn]"
            ]
        }
        self.special_tokens = special_tokens
        self.tokenizer = AutoTokenizer.from_pretrained(token_path)
        self.tokenizer.add_special_tokens(special_tokens)
        self.skip_special_tokens = skip_special_tokens

    def encode(self, text, **kwargs):
        tokens = self.tokenizer([text], return_tensors="pt")
        tokens = tokens["input_ids"][0].cpu().tolist()
        return tokens

    def decode(self, tokens):
        tokens = torch.tensor(tokens, dtype=torch.int64)
        text = self.tokenizer.batch_decode([tokens], skip_special_tokens=self.skip_special_tokens)[0]
        return text


@lru_cache(maxsize=None)
def get_qwen_tokenizer(
    token_path: str,
    skip_special_tokens: bool
) -> QwenTokenizer:
    return QwenTokenizer(token_path=token_path, skip_special_tokens=skip_special_tokens)



import datetime
import logging
import uuid

from pydantic import BaseModel

try:
    from livekit import api
except ModuleNotFoundError as e:
    logging.error(f"Exception: {e}")
    logging.error("In order to use Livekit API, you need to `pip install achatbot[livekit-api]`.")
    raise Exception(f"Missing module: {e}")

from src.common.const import *
from src.common.types import GeneralRoomInfo, LivekitRoomArgs
from src.common.interface import IRoomManager
from src.common.factory import EngineClass


from dotenv import load_dotenv

load_dotenv(override=True)


class LivekitRoom(EngineClass, IRoomManager):
    TAG = "livekit_room"

    def __init__(self, **kwargs) -> None:
        self.args = LivekitRoomArgs(**kwargs)
        # live kit http api with pb(protobuf) serialize data
        self._http_api = api.LiveKitAPI() if self.args.is_common_session else None

    async def close_session(self):
        if self._http_api:
            await self._http_api.aclose()

    async def create_room(
        self, room_name: str | None = None, exp_time_s: int = ROOM_EXPIRE_TIME
    ) -> GeneralRoomInfo:
        name = room_name
        empty_time_s = exp_time_s if exp_time_s else ROOM_EXPIRE_TIME
        if not room_name:
            name = "room_" + str(uuid.uuid4())
            if not exp_time_s:
                empty_time_s = RANDOM_ROOM_EXPIRE_TIME

        http_api = self._http_api if self._http_api else api.LiveKitAPI()
        room = await http_api.room.create_room(
            api.CreateRoomRequest(
                name=name,
                empty_timeout=empty_time_s,
            )
        )
        logging.debug(f"create_room:{room}")

        if not self._http_api:
            await http_api.aclose()

        g_room = GeneralRoomInfo(
            sid=room.sid,
            name=room.name,
            ttl_s=room.empty_timeout,
            creation_time=str(room.creation_time),
            # NOTE: room.__dict__, python3.10 don't ok,python.3.11 ok
            extra_data={
                "enabled_codecs": room.enabled_codecs,
                "departure_timeout": room.departure_timeout,
                "max_participants": room.max_participants,
                "num_participants": room.num_participants,
                "num_publishers": room.num_publishers,
                "active_recording": room.active_recording,
                "version": room.version,
            },
        )

        return g_room

    async def gen_token(self, room_name: str, exp_time_s: int = ROOM_TOKEN_EXPIRE_TIME) -> str:
        user_identity = str(uuid.uuid4())
        token = (
            api.AccessToken()
            .with_identity(user_identity)
            .with_name(self.args.bot_name)
            .with_grants(
                api.VideoGrants(
                    room_join=True,
                    room=room_name,
                )
            )
            .with_ttl(datetime.timedelta(seconds=exp_time_s))
            .to_jwt()
        )
        logging.debug(f"token:{token}")

        return token

    async def get_room(self, room_name: str) -> GeneralRoomInfo:
        http_api = self._http_api if self._http_api else api.LiveKitAPI()
        result = await http_api.room.list_rooms(api.ListRoomsRequest(names=[room_name]))

        logging.debug(f"list_rooms:{result.rooms}")
        if len(result.rooms) == 0:
            return None

        if not self._http_api:
            await http_api.aclose()

        room = result.rooms[0]
        g_room = GeneralRoomInfo(
            sid=room.sid,
            name=room.name,
            creation_time=str(room.creation_time),
            extra_data=room.__dict__,
        )

        return g_room

    async def check_valid_room(self, room_name: str, token: str) -> bool:
        if not room_name or not token:
            return False

        try:
            api.TokenVerifier().verify(token)
        except Exception as e:
            logging.warning(f"{token} verify Exception: {e}")
            return False

        room = await self.get_room(room_name)
        if not room:
            return False

        return True



import logging

from achatbot.common.register import Register


class FunctionManager:
    """
    FunctionManager just as a namespace to use static methods.
    """

    functions = Register("llm_function_calling")

    def __init__(self):
        raise RuntimeError("FunctionManager is not intended to be instantiated")

    @staticmethod
    def get_tool_calls() -> list[dict]:
        tool_calls = []
        for _, func_cls in FunctionManager.functions.items():
            tool_calls.append(func_cls.get_tool_call())
        return tool_calls

    @staticmethod
    def execute(name: str, session, **args):
        func_cls = FunctionManager.functions[name]
        return func_cls.execute(session, **args)



import logging
import os

from src.common.types import MODELS_DIR
from src.common import interface
from src.common.factory import EngineClass, EngineFactory

from dotenv import load_dotenv

load_dotenv(override=True)


class ASREnvInit:
    @staticmethod
    def getEngine(tag, **kwargs) -> interface.IAsr | EngineClass:
        if "sense_voice" in tag:
            from . import sense_voice_asr
        elif "groq" in tag:
            from . import whisper_groq_asr
        elif "whisper_" in tag:
            from . import whisper_asr

        engine = EngineFactory.get_engine_by_tag(EngineClass, tag, **kwargs)
        return engine

    @staticmethod
    def initASREngine(
        tag: str | None = None,
        # kwargs: dict | None = None
    ) -> interface.IAsr | EngineClass:
        # asr
        tag = tag or os.getenv("ASR_TAG", "whisper_timestamped_asr")
        kwargs = {}
        kwargs["model_name_or_path"] = os.getenv("ASR_MODEL_NAME_OR_PATH", "base")
        kwargs["download_path"] = MODELS_DIR
        kwargs["verbose"] = bool(os.getenv("ASR_VERBOSE", "True"))
        kwargs["language"] = os.getenv("ASR_LANG", "zh")
        engine = ASREnvInit.getEngine(tag, **kwargs)
        logging.info(f"initASREngine: {tag}, {engine}")
        return engine



from dataclasses import dataclass, field


@dataclass
class TransformersGoTOCRArgs:
    r"""
    just use transformers Got OCR2.0
    """

    lm_model_name_or_path: str = field(
        # see: https://huggingface.co/stepfun-ai/GOT-OCR2_0/discussions/26
        default="weege007/GOT-OCR2_0",  # from stepfun-ai/GOT-OCR2_0, change modeling_GOT.py for streamer
        metadata={
            "help": "The pretrained language model to use. Default is 'weege007/GOT-OCR2_0'."
        },
    )
    lm_device_map: str | dict | None = field(
        default=None,
        metadata={
            "help": "The device map for multi cpu/gpu. use 'cuda' for GPU acceleration, 'mps','cpu','auto', dict, None. default None"
        },
    )
    lm_device: str = field(
        default="cpu",
        metadata={
            "help": "The device for single cpu/mps/gpu. use 'cuda' for GPU acceleration, 'mps'(apple), 'cpu'. default cpu"
        },
    )
    lm_torch_dtype: str = field(
        default="bfloat16",
        metadata={
            "help": "The PyTorch data type for the model and input tensors. One of `float32` (full-precision), `float16` or `bfloat16` (both half-precision). default bfloat16"
        },
    )
    # https://huggingface.co/docs/transformers/perf_infer_gpu_one
    lm_attn_impl: str | None = field(
        default=None,
        metadata={
            "help": "The attention implementation to use. One of 'sdpa', 'flash_attention_2', default no attention implementation."
        },
    )
    warnup_prompt: str = field(
        default="Repeat the word 'weedge niu bi'.",
        metadata={"help": "warnup llm generate prompt. Default is 'weedge niu bi'."},
    )
    warnup_steps: int = field(
        default=2,
        metadata={"help": "The number of steps to run the warmup prompt. Default is 2."},
    )
    lm_stream: bool = field(
        default=True,
        metadata={
            "help": "Whether to use streaming; set this to True for streaming output. Default is True."
        },
    )
    model_type: str = field(
        default="chat",
        metadata={"help": "Model type, chat, chat_crop. Default is chat."},
    )
    ocr_type: str = field(
        default="ocr",
        metadata={"help": "OCR type, ocr,format. Default is ocr."},
    )
    ocr_box: str = field(
        default="",
        metadata={
            "help": "ocr_box bbox, e.g.:'[x1,y1]' or '[x1,y1,x2,y2]' . Default is empty str."
        },
    )
    ocr_color: str = field(
        default="",
        metadata={"help": "ocr_color, e.g.:'GREEN'. Default is empty str."},
    )
    conv_mode: str = field(
        default="mpt",
        metadata={"help": "conversation mode, e.g.:'mpt'. Default is 'mpt'."},
    )
    render: bool = field(
        default=False,
        metadata={"help": "Whether to render to file, Default is false."},
    )
    save_render_file: str = field(
        default="",
        metadata={"help": "when render is true, to save file path, Default is empty str."},
    )



from concurrent import futures
import logging
import os
import sys

import grpc
from dotenv import load_dotenv

try:
    cur_dir = os.path.dirname(__file__)
    sys.path.insert(1, os.path.join(cur_dir, "../../../../common/grpc/idl"))
    from achatbot.common.grpc.idl.tts_pb2_grpc import add_TTSServicer_to_server
except ModuleNotFoundError as e:
    raise Exception(f"grpc import error: {e}")

from achatbot.common.grpc.interceptors.authentication_server import AuthenticationInterceptor
from achatbot.cmd.grpc.speaker.server.servicers.tts import TTS
from achatbot.common.logger import Logger

load_dotenv(override=True)

Logger.init(
    os.getenv("LOG_LEVEL", "debug").upper(),
    app_name="chat-bot-tts-serve",
    is_file=False,
    is_console=True,
)


def serve() -> None:
    port = os.getenv("PORT", "50052")
    max_workers = int(os.getenv("MAX_WORKERS", "10"))
    logging.info(f"serve port: {port} max_workers: {max_workers}")
    token = "oligei-tts"
    authenticator = AuthenticationInterceptor(
        "authorization", token, grpc.StatusCode.UNAUTHENTICATED, "Access denied!"
    )
    server = grpc.server(
        futures.ThreadPoolExecutor(max_workers=max_workers),
        interceptors=(authenticator,),
    )
    add_TTSServicer_to_server(TTS(), server)
    server.add_insecure_port(f"[::]:{port}")
    server.start()
    logging.info(f"Server started port: {port}")
    server.wait_for_termination()


# python -m achatbot.cmd.grpc.speaker.server.serve
if __name__ == "__main__":
    serve()



import modal
import os


class ContainerRuntimeConfig:
    images = {
        "default": (
            modal.Image.debian_slim(python_version="3.11")
            .apt_install("git", "git-lfs", "ffmpeg")
            .pip_install(
                [
                    "achatbot["
                    "fastapi_bot_server,"
                    "websocket_server_transport,"
                    "silero_vad_analyzer,"
                    "moshi_voice_processor"
                    "]~=0.0.8.3",
                    "huggingface_hub[hf_transfer]==0.24.7",
                ],
                extra_index_url="https://pypi.org/simple/",
            )
            .env(
                {
                    "HF_HUB_ENABLE_HF_TRANSFER": "1",
                    "ACHATBOT_PKG": "1",
                    "BOT_CONFIG_NAME": os.getenv(
                        "BOT_CONFIG_NAME", "fastapi_websocket_moshi_voice_bot"
                    ),
                    "LOG_LEVEL": os.getenv("LOG_LEVEL", "info"),
                    "IMAGE_NAME": os.getenv("IMAGE_NAME", "default"),
                }
            )
        ),
    }

    @staticmethod
    def get_img(image_name: str = None):
        image_name = image_name or os.getenv("IMAGE_NAME", "default")
        if image_name not in ContainerRuntimeConfig.images:
            raise Exception(f"image name {image_name} not found")
        print(f"use image:{image_name}")
        return ContainerRuntimeConfig.images[image_name]


img = ContainerRuntimeConfig.get_img()
with img.imports():
    import asyncio
    import logging
    import pathlib
    import os

    from fastapi import WebSocket
    from huggingface_hub import hf_hub_download
    from moshi.models import loaders

    from achatbot.cmd.bots.base_fastapi_websocket_server import AIFastapiWebsocketBot
    from achatbot.cmd.http.server.fastapi_daily_bot_serve import app as fastapi_app
    from achatbot.cmd.bots.bot_loader import BotLoader
    from achatbot.common.logger import Logger

# ----------------------- app -------------------------------
app = modal.App("fastapi_ws_moshi_voice_bot")

volume = modal.Volume.from_name("bot_config", create_if_missing=True)


# 128 MiB of memory and 0.125 CPU cores by default container runtime
@app.cls(
    image=img,
    # secrets=[modal.Secret.from_name("achatbot")],
    volumes={"/bots": volume},
    gpu="A10G",
    container_idle_timeout=1200,
    timeout=600,
    allow_concurrent_inputs=1,
)
class Srv:
    @modal.build()
    def setup(self):
        Logger.init(os.getenv("LOG_LEVEL", "info").upper(), is_file=False, is_console=True)
        # https://huggingface.co/docs/huggingface_hub/guides/download
        repo_id = loaders.DEFAULT_REPO
        bot_config_name = os.getenv("BOT_CONFIG_NAME", "")
        if "moshika" in bot_config_name:
            repo_id = "kyutai/moshika-pytorch-bf16"
        hf_hub_download(repo_id, loaders.MOSHI_NAME)
        hf_hub_download(repo_id, loaders.MIMI_NAME)
        hf_hub_download(repo_id, loaders.TEXT_TOKENIZER_NAME)
        logging.info(f"download model done, config name:{bot_config_name}, hf_repo:{repo_id}")

    @modal.enter()
    def enter(self):
        volume.reload()
        bot_config_name = os.getenv("BOT_CONFIG_NAME", "")
        self.config_path = pathlib.Path(os.path.join("/bots", f"{bot_config_name}.json"))
        # self.run_bot: AIFastapiWebsocketBot = asyncio.run(
        #    BotLoader.load_bot(self.config_path, bot_type="fastapi_ws_bot"))
        logging.info(f"start enter get run_bot from conifg:{self.config_path}")

    @modal.asgi_app()
    def app(self):
        @fastapi_app.websocket("/")
        async def websocket_endpoint(websocket: WebSocket):
            self.run_bot: AIFastapiWebsocketBot = await BotLoader.load_bot(  # type: ignore
                self.config_path, bot_type="fastapi_ws_bot"
            )

            # NOTE: after init, websocket to accept connection, then to run
            await websocket.accept()
            self.run_bot.set_fastapi_websocket(websocket)
            logging.info(f"accept client: {websocket.client}")
            await self.run_bot.try_run()

        return fastapi_app



import os
import logging


def get_tidb_url():
    url = f"mysql+pymysql://{os.getenv('TIDB_USERNAME')}:{os.getenv('TIDB_PASSWORD')}@{os.getenv('TIDB_HOST')}:{os.getenv('TIDB_PORT')}/{os.getenv('TIDB_DATABASE')}?ssl_ca={os.getenv('TIDB_SSL_CA')}&ssl_verify_cert=true&ssl_verify_identity=true"
    logging.info(f"tidb url: {url}")
    return url



import os
import logging
import sys
from typing import AsyncGenerator, Generator
import uuid

import torch
from transformers import WhisperFeatureExtractor, AutoTokenizer
from apipeline.frames import *

from deps.GLM4Voice.speech_tokenizer.modeling_whisper import WhisperVQEncoder
from deps.GLM4Voice.audio_process import AudioStreamProcessor
from deps.GLM4Voice.flow_inference import AudioDecoder
from deps.GLM4Voice.speech_tokenizer.utils import extract_speech_token

from src.core.llm.transformers.manual_voice_glm import TransformersManualVoicGLM
from src.processors.voice.base import VoiceProcessorBase
from src.types.llm.lmgen import *
from src.types.llm.transformers import TransformersLMArgs
from src.common.session import Session
from src.common.types import SessionCtx
from src.common.utils.audio_utils import bytes2TorchTensorWith16
from src.types.frames import *


class GLMVoiceBaseProcessor(VoiceProcessorBase):
    """
    use Voice-Tokenizer (whisper) +  GLM4-Voice-9B(text/audio) + Voice-Decoder (CosyVoice)
    - T1A1-T2A2: (text/speech)-to-(tokens) (GLM4(text)/whisper(speech encoder)) -> GLM4(llm) -- text|speech tokens --> GLM4(text decoder)|CosyVoice(speech decoder(mel->waveform)))

    Model Architecture:
    - GLM-4-Voice-Tokenizer: Trained by adding vector quantization to the encoder part of Whisper, converting continuous speech input into discrete tokens.  Each second of audio is converted into 12.5 discrete tokens.
    - GLM-4-Voice-9B: Pre-trained and aligned on speech modality based on GLM-4-9B, enabling understanding and generation of discretized speech.
    - GLM-4-Voice-Decoder: A speech decoder supporting streaming inference, retrained based on CosyVoice, converting discrete speech tokens into continuous speech output. Generation can start with as few as 10 audio tokens, reducing conversation latency.
    """

    def __init__(
        self,
        *,
        voice_in_args: GLMVoiceInArgs | dict = GLMVoiceInArgs(),
        lm_gen_args: GLMInferenceArgs | dict = GLMInferenceArgs(),
        voice_out_args: GLMVoiceOutArgs | dict = GLMVoiceOutArgs(),
        system_prompt: str = "",
        voice_tokenizer_path: str | None = None,  # audio encoder/ft extractor
        model_path: str | None = None,  # gen lm and text tokenizer
        voice_decoder_path: str | None = None,  # audio decoder
        device: str = "cuda",
        torch_dtype: str = "auto",  # auto,float16,bfloat16,float32
        bnb_quant_type: str = "int4",
        session: Session | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        cur_dir = os.path.dirname(__file__)
        if bool(os.getenv("ACHATBOT_PKG", "")):
            sys.path.insert(1, os.path.join(cur_dir, "../../GLM4Voice"))
            sys.path.insert(2, os.path.join(cur_dir, "../../GLM4Voice/third_party/Matcha-TTS"))
        else:
            sys.path.insert(1, os.path.join(cur_dir, "../../../deps/GLM4Voice"))
            sys.path.insert(
                2, os.path.join(cur_dir, "../../../deps/GLM4Voice/third_party/Matcha-TTS")
            )

        self._voice_in_args = voice_in_args
        if isinstance(voice_in_args, dict):
            self._voice_in_args = GLMVoiceInArgs(**voice_in_args)
        self._lm_gen_args = lm_gen_args
        if isinstance(lm_gen_args, dict):
            self._lm_gen_args = GLMInferenceArgs(**lm_gen_args)
        self._voice_out_args = voice_out_args
        if isinstance(voice_out_args, dict):
            self._voice_out_args = GLMVoiceOutArgs(**voice_out_args)

        self._sys_prompt = system_prompt or TransformersManualVoicGLM.DEFAULT_SYS_PROMPT
        self._voice_tokenizer_path = voice_tokenizer_path
        self._model_path = model_path
        self._voice_decoder_path = voice_decoder_path
        self._torch_dtype = torch_dtype
        self._bnb_quant_type = bnb_quant_type
        self._device = device

        self._session = session or Session(**SessionCtx(uuid.uuid4()).__dict__)

        self.reset()
        self.load_models()

        # "你好!有什么可以帮你的吗？" text tokens + audio waveform tokens
        self.test_tokens = [
            109377,
            6313,
            101665,
            109213,
            99215,
            99444,
            99212,
            11314,
            166648,
            164526,
            163431,
            155030,
            160561,
            167785,
            165922,
            167515,
            163387,
            157845,
            157845,
            158729,
            159316,
            153554,
            158936,
            160466,
            157871,
            168288,
            165890,
            152620,
            164640,
            158589,
            152703,
            154548,
            165926,
            163830,
            166537,
            165003,
            155773,
            161696,
            158134,
            167188,
            161032,
            157635,
            164078,
            160166,
            160014,
            160014,
            160014,
            166768,
            151336,
        ]

    @property
    def stream_info(self) -> dict:
        """Return dict out stream info"""
        return {
            "sample_rate": self._voice_out_args.audio_sample_rate,
            "channels": self._voice_out_args.audio_channels,
        }

    def reset(self):
        # input_texts + completion_texts (audio tokenid special tag)
        self._history_texts = ""

    def load_models(self):
        logging.info("loading model weights")

        # Speech tokenizer(whisper vq encoder)/feature_extractor
        self._whisper_model = (
            WhisperVQEncoder.from_pretrained(self._voice_tokenizer_path).eval().to(self._device)
        )
        self._feature_extractor = WhisperFeatureExtractor.from_pretrained(
            self._voice_tokenizer_path
        )
        logging.info("speech whisper vq encoder and feature_extractor model state weight load")

        # Flow & Hift decoder with config, fixed sample rate 22050
        flow_config = os.path.join(self._voice_decoder_path, "config.yaml")
        flow_checkpoint = os.path.join(self._voice_decoder_path, "flow.pt")
        hift_checkpoint = os.path.join(self._voice_decoder_path, "hift.pt")
        self._audio_decoder = AudioDecoder(
            config_path=flow_config,
            flow_ckpt_path=flow_checkpoint,
            hift_ckpt_path=hift_checkpoint,
            device=self._device,
        )
        logging.info("speech audio Flow & Hift decoder model state weight load")

        # gen lm text tokenizer
        self._glm_tokenizer = AutoTokenizer.from_pretrained(
            self._model_path, trust_remote_code=True
        )
        self._audio_offset = self._glm_tokenizer.convert_tokens_to_ids("<|audio_0|>")
        self._end_token_id = self._glm_tokenizer.convert_tokens_to_ids("<|user|>")
        logging.info(
            f"gen lm text tokenizer load,audio_offset:{self._audio_offset} end_token_id:{self._end_token_id}"
        )

        # gen lm
        self._glm_model = TransformersManualVoicGLM(
            **TransformersLMArgs(
                lm_model_name_or_path=self._model_path,
                lm_gen_temperature=self._lm_gen_args.temperature,
                lm_gen_top_p=self._lm_gen_args.top_p,
                lm_gen_max_new_tokens=self._lm_gen_args.max_new_token,
                lm_torch_dtype=self._torch_dtype,
                lm_bnb_quant_type=self._bnb_quant_type,
                lm_device=self._device,
            ).__dict__
        )
        self._glm_model.warmup()
        logging.info("gen lm model state weight load and warnup")

        logging.info("model weights loaded")

    async def start(self, frame: StartFrame):
        await super().start(frame)

        self._create_push_task()

        logging.info("start done")

    async def stop(self, frame: EndFrame):
        await super().stop(frame)
        logging.info("stop done")

    async def cancel(self, frame: CancelFrame):
        await super().cancel(frame)
        logging.info("cancel done")

    async def run_voice(self, frame: AudioRawFrame) -> AsyncGenerator[Frame, None]:
        if isinstance(frame, PathAudioRawFrame):
            utt = frame.path
        else:
            audio_tensor = bytes2TorchTensorWith16(frame.audio)
            utt = (audio_tensor, self._voice_in_args.audio_sample_rate)
        # default 16khz sample rate encode/extract -> audio tokens
        audio_tokens = extract_speech_token(self._whisper_model, self._feature_extractor, [utt])[0]
        if len(audio_tokens) == 0:
            yield ErrorFrame("No audio tokens extracted")
            return

        audio_tokens = "".join([f"<|audio_{x}|>" for x in audio_tokens])
        audio_tokens = "<|begin_of_audio|>" + audio_tokens + "<|end_of_audio|>"
        user_input = audio_tokens

        # history
        if "<|system|>" not in self._history_texts:
            self._history_texts += f"<|system|>\n{self._sys_prompt}"
        self._history_texts += f"<|user|>\n{user_input}<|assistant|>streaming_transcription\n"
        # TODO: check token max length

        self._session.ctx.state["prompt"] = self._history_texts
        # await self.tokens_decode_out(self.test_tokens)
        # u can mock GLM generate for dev
        iter_tokens = self._glm_model.generate(self._session)
        await self.tokens_decode_out(iter_tokens)

        yield None

    @torch.no_grad()
    async def tokens_decode_out(self, iter_tokens):
        text_tokens, audio_tokens = [], []
        complete_tokens = []

        prompt_speech_feat = torch.zeros(1, 0, 80).to(self._device)
        flow_prompt_speech_token = torch.zeros(1, 0, dtype=torch.int64).to(self._device)

        this_uuid = str(uuid.uuid4())
        tts_speechs = []
        tts_mels = []
        prev_mel = None

        is_finalize = False
        block_size_list = [25, 50, 100, 150, 200]
        block_size_idx = 0
        block_size = block_size_list[block_size_idx]

        # default 22050 hz sample rate, match with audio_decoder 22050 hz
        audio_processor = AudioStreamProcessor(sr=self._voice_out_args.audio_sample_rate)
        for token_id in iter_tokens:
            if token_id == self._end_token_id:
                is_finalize = True
            if len(audio_tokens) >= block_size or (is_finalize and audio_tokens):
                if block_size_idx < len(block_size_list) - 1:
                    block_size_idx += 1
                    block_size = block_size_list[block_size_idx]
                tts_token = torch.tensor(audio_tokens, device=self._device).unsqueeze(0)

                if prev_mel is not None:
                    prompt_speech_feat = torch.cat(tts_mels, dim=-1).transpose(1, 2)

                # gen waveform and mel-spectrogram feat
                tts_speech, tts_mel = self._audio_decoder.token2wav(
                    tts_token,
                    uuid=this_uuid,
                    prompt_token=flow_prompt_speech_token.to(self._device),
                    prompt_feat=prompt_speech_feat.to(self._device),
                    finalize=is_finalize,
                )
                prev_mel = tts_mel

                # waveform tensor write to audio bytes
                audio_bytes = audio_processor.process(
                    tts_speech.clone().cpu().numpy()[0], last=is_finalize
                )

                tts_speechs.append(tts_speech.squeeze())
                tts_mels.append(tts_mel)
                if audio_bytes:
                    # print("audio_bytes====>", len(audio_bytes))
                    await self.queue_frame(
                        AudioRawFrame(
                            audio=audio_bytes,
                            sample_rate=self._voice_out_args.audio_sample_rate,
                            num_channels=self._voice_out_args.audio_channels,
                        )
                    )
                flow_prompt_speech_token = torch.cat((flow_prompt_speech_token, tts_token), dim=-1)
                audio_tokens = []

            if not is_finalize:
                complete_tokens.append(token_id)
                if token_id >= self._audio_offset:
                    audio_tokens.append(token_id - self._audio_offset)
                else:
                    text = self._glm_tokenizer.decode([token_id], ignore_special_tokens=False)
                    if text:
                        await self.push_frame(TextFrame(text))
                    text_tokens.append(token_id)

        complete_text = self._glm_tokenizer.decode(
            complete_tokens, spaces_between_special_tokens=False
        )
        self._history_texts += complete_text
        logging.info(f"history_texts:{self._history_texts}")


class GLMAudioVoiceProcessor(GLMVoiceBaseProcessor):
    """
    use Voice-Tokenizer (whisper) +  GLM4-Voice-9B(text/audio) + Text-Tokenizer (GLM-tokenizer), Voice-Decoder (CosyVoice)
    - A1-T2A2: (text/speech)-to-(tokens) (GLM4(text)/whisper(speech encoder)) -> GLM4(llm) -- text|speech tokens --> GLM4(text decoder)|CosyVoice(speech decoder(mel->waveform)))
    """

    pass


class GLMTextVoiceProcessor(GLMVoiceBaseProcessor):
    """
    use Text-Tokenizer (GLM-tokenizer) +  GLM4-Voice-9B(text/audio) + Text-Tokenizer (GLM-tokenizer), Voice-Decoder (CosyVoice)
    - T1/A1-T2A2: (text/speech)-to-(tokens) (GLM4(text)/whisper(speech encoder)) -> GLM4(llm) -- text|speech tokens --> GLM4(text decoder)|CosyVoice(speech decoder(mel->waveform)))
    """

    async def run_text(self, frame: TextFrame) -> AsyncGenerator[Frame, None]:
        user_input = frame.text.strip()
        # history
        if "<|system|>" not in self._history_texts:
            self._history_texts += f"<|system|>\n{self._sys_prompt}"
        self._history_texts += f"<|user|>\n{user_input}<|assistant|>streaming_transcription\n"

        self._session.ctx.state["prompt"] = self._history_texts
        iter_tokens = self._glm_model.generate(self._session)

        await self.tokens_decode_out(iter_tokens)

        yield None
